from unittest import TestCase

import remotior_sensus


class TestSpectralSignatures(TestCase):

    def test_spectral_signatures(self):
        rs = remotior_sensus.Session(
            n_processes=2, available_ram=1000, log_level=10
            )
        cfg = rs.configurations
        cfg.logger.log.debug('test')
        # create Spectral Signature Catalog
        cfg.logger.log.debug('>>> test create Spectral Signature Catalog')
        signature_catalog = rs.spectral_signatures_catalog()
        value_list = [1.1, 1.2, 1.3, 1.4]
        wavelength_list = [1, 2, 3, 4]
        signature_catalog.add_spectral_signature(
            value_list=value_list, wavelength_list=wavelength_list
            )
        self.assertEqual(len(signature_catalog.signatures), 1)
        signature_catalog.add_spectral_signature(
            value_list=value_list, macroclass_id=2,
            wavelength_list=wavelength_list
            )
        self.assertEqual(signature_catalog.table.shape[0], 2)
        cfg.logger.log.debug('>>> test save Spectral Signature Catalog')
        temp = cfg.temp.temporary_file_path(name_suffix='.scpx')
        signature_catalog.save(output_path=temp)
        self.assertTrue(rs.files_directories.is_file(temp))
        signature_catalog_x = rs.spectral_signatures_catalog()
        signature_catalog_x.load(file_path=temp)
        self.assertEqual(signature_catalog.table.shape[0],
                         signature_catalog_x.table.shape[0])
        self.assertEqual(signature_catalog.table[0],
                         signature_catalog_x.table[0])
        self.assertEqual(len(signature_catalog.signatures),
                         len(signature_catalog_x.signatures))
        # create BandSet
        cfg.logger.log.debug('>>> test create BandSet')
        catalog = rs.bandset_catalog()
        file_list = ['S2_2020-01-01/S2_B02.tif', 'S2_2020-01-01/S2_B03.tif',
                     'S2_2020-01-01/S2_B04.tif', 'S2_2020-01-01/S2_B05.tif']
        root_directory = 'data'
        catalog.create_bandset(
            file_list, wavelengths=['Sentinel-2'],
            root_directory=root_directory
            )
        # set BandSet in SpectralCatalog
        signature_catalog_2 = rs.spectral_signatures_catalog(
            bandset=catalog.get(1)
        )
        # add spectral signature getting wavelength from BandSet
        signature_catalog_2.add_spectral_signature(value_list=value_list)
        self.assertEqual(
            signature_catalog_2.signatures[
                list(signature_catalog_2.signatures.keys())[0]][
                'wavelength'].tolist(), catalog.get(1).get_wavelengths()
        )
        # add spectral signature with macroclass_id = 2
        signature_catalog_2.add_spectral_signature(
            value_list=value_list, macroclass_id=2
            )
        self.assertEqual(
            signature_catalog_2.table[
                signature_catalog_2.table[
                    'macroclass_id'] == 2].macroclass_id, 2
        )
        # import spectral signature csv
        cfg.logger.log.debug('>>> test import spectral signature csv')
        # signature with standard deviation
        signature_catalog_2.import_spectral_signature_csv(
            csv_path='data/files/spectral_signature_1.csv', macroclass_id=4,
            class_id=4, macroclass_name='imported_1', class_name='imported_1'
        )
        self.assertEqual(
            signature_catalog_2.table[
                signature_catalog_2.table['macroclass_id'] == 4].macroclass_id,
            4
        )
        # signature without standard deviation
        signature_catalog_2.import_spectral_signature_csv(
            csv_path='data/files/spectral_signature_2.csv', macroclass_id=5,
            class_id=5, macroclass_name='imported_2', class_name='imported_2'
        )
        self.assertEqual(
            signature_catalog_2.table[
                signature_catalog_2.table[
                    'macroclass_id'] == 5].macroclass_id,
            5
            )
        # import vector
        cfg.logger.log.debug('>>> test import vector')
        signature_catalog_2.import_vector(
            file_path='data/files/roi.gpkg', macroclass_value=7,
            class_field='class', macroclass_name_field='macroclass',
            class_name_field='class', calculate_signature=True
        )
        self.assertEqual(
            signature_catalog_2.table[
                signature_catalog_2.table['macroclass_id'] == 4].macroclass_id,
            4
        )
        # import vector
        cfg.logger.log.debug('>>> test remove signature')
        temp = cfg.temp.temporary_file_path(name_suffix='.scpx')
        signature_catalog_2.save(output_path=temp)
        self.assertTrue(rs.files_directories.is_file(temp))
        sig_id = signature_catalog_2.table[
            signature_catalog_2.table['macroclass_id'] == 7].signature_id[0]
        signature_catalog_2.remove_signature_by_id(signature_id=sig_id)
        temp = cfg.temp.temporary_file_path(name_suffix='.scpx')
        signature_catalog_2.save(output_path=temp)
        self.assertTrue(rs.files_directories.is_file(temp))
        shape = signature_catalog_2.table.shape[0]
        signature_catalog_2.import_file(file_path=temp)
        self.assertEqual(
            signature_catalog_2.table.shape[0],
            len(signature_catalog_2.signatures)
        )
        self.assertEqual(
            signature_catalog_2.table.shape[0], shape * 2
        )
        # region growing
        cfg.logger.log.debug('>>> test region growing')
        catalog_3 = rs.bandset_catalog()
        file_list = ['S2_2020-01-04/S2_B02.tif', 'S2_2020-01-04/S2_B03.tif',
                     'S2_2020-01-04/S2_B04.tif']
        root_directory = 'data'
        catalog_3.create_bandset(
            file_list, wavelengths=['Sentinel-2'],
            root_directory=root_directory
            )
        region_vector = rs.shared_tools.region_growing_polygon(
            coordinate_x=230303, coordinate_y=4674704,
            input_bands=1, max_width=4,
            max_spectral_distance=0.1, minimum_size=1,
            bandset_catalog=catalog_3
        )
        self.assertTrue(rs.files_directories.is_file(region_vector))
        # set BandSet in SpectralCatalog
        signature_catalog_3 = rs.spectral_signatures_catalog(
            bandset=catalog_3.get(1)
        )
        signature_catalog_3.import_vector(
            file_path=region_vector, macroclass_value=5,
            class_value=11, macroclass_name='macroclass',
            class_name='class', calculate_signature=True
        )
        self.assertEqual(
            signature_catalog_3.table[
                signature_catalog_3.table['macroclass_id'] == 5].macroclass_id,
            5
        )
        region_vector = rs.shared_tools.region_growing_polygon(
            coordinate_x=230316, coordinate_y=4674708,
            input_bands=1, max_width=4,
            max_spectral_distance=0.1, minimum_size=1,
            bandset_catalog=catalog_3
        )
        signature_catalog_3.import_vector(
            file_path=region_vector, macroclass_value=7,
            class_value=21, macroclass_name='macroclass_2',
            class_name='class_2', calculate_signature=True
        )
        signature_7_pixel_count = signature_catalog_3.table[
            signature_catalog_3.table['macroclass_id'] == 7].pixel_count[0]
        signature_7_unit = signature_catalog_3.table[
            signature_catalog_3.table['macroclass_id'] == 7].unit[0]
        self.assertTrue(signature_7_pixel_count > 0)
        self.assertEqual(signature_7_unit, cfg.wl_micro)
        signature_count = signature_catalog_3.table.shape[0]
        ids = signature_catalog_3.table.signature_id.tolist()
        signature_catalog_3.merge_signatures_by_id(
            signature_id_list=ids, calculate_signature=False)
        self.assertEqual(
            signature_catalog_3.table.shape[0], signature_count + 1
        )
        signature = signature_catalog_3.table[
            signature_catalog_3.table['signature'] == 0].signature_id[0]
        ids.append(signature)
        signature_catalog_3.merge_signatures_by_id(
            signature_id_list=ids, calculate_signature=True,
            class_name='merged3'
        )
        self.assertEqual(
            signature_catalog_3.table[
                signature_catalog_3.table['class_name'] == 'merged3'
            ].signature, 1
        )
        # signature distance
        cfg.logger.log.debug('>>> test signature distance')
        # calculate Bray-Curtis similarity
        bray_curtis = signature_catalog_3.calculate_bray_curtis_similarity(
            signature_id_x=ids[0], signature_id_y=ids[1]
        )
        self.assertTrue(bray_curtis > 0)
        # calculate Euclidean Distance
        euclidean_distance = signature_catalog_3.calculate_euclidean_distance(
            signature_id_x=ids[0], signature_id_y=ids[1]
        )
        self.assertTrue(euclidean_distance > 0)
        # calculate spectral angle
        spectral_angle = signature_catalog_3.calculate_spectral_angle(
            signature_id_x=ids[0], signature_id_y=ids[1]
        )
        self.assertTrue(spectral_angle > 0)
        # signature to plot
        cfg.logger.log.debug('>>> test signature to plot')
        plot_catalog = rs.spectral_signatures_plot_catalog()
        signature_id = signature_catalog_3.table.signature_id[0]
        signature_catalog_3.export_signature_values_for_plot(
            signature_id=signature_id, plot_catalog=plot_catalog
            )
        self.assertEqual(plot_catalog.get_signature_count(), 1)
        for plot in plot_catalog.catalog:
            signature = plot_catalog.get_signature(signature_id=plot)

            self.assertEqual(signature.signature_id, signature_id)
        signature_ids = plot_catalog.get_signature_ids()
        self.assertEqual(len(signature_ids), 1)
        """
        # interactive plot
        signature_catalog_3.add_signatures_to_plot_by_id(signature_ids)
        """
        plot_catalog.remove_signature(signature_id)
        self.assertEqual(plot_catalog.get_signature_count(), 0)
        """
        # interactive plot
        histogram = signature_catalog_3.calculate_scatter_plot_by_id(
            signature_id=signature_id, band_x=1, band_y=2, decimal_round=1,
            plot=True
        )
        """
        histogram = signature_catalog_3.calculate_scatter_plot_by_id(
            signature_id=signature_id, band_x=1, band_y=2, decimal_round=1
        )
        self.assertTrue(histogram is not None)

        # export vector
        temp_gpkg = cfg.temp.temporary_file_path(name_suffix='.gpkg')
        signature_catalog_3.export_vector(
            signature_id_list=[signature_id], output_path=temp_gpkg,
            vector_format='GPKG'
        )
        self.assertTrue(rs.files_directories.is_file(temp_gpkg))
        temp_shp = cfg.temp.temporary_file_path(name_suffix='.shp')
        signature_catalog_3.export_vector(
            signature_id_list=[signature_id], output_path=temp_shp,
            vector_format='ESRI Shapefile'
        )
        self.assertTrue(rs.files_directories.is_file(temp_shp))
        # export as csv
        csv = signature_catalog_3.export_signatures_as_csv(
            signature_id_list=[signature_id], output_directory=cfg.temp.dir
        )
        self.assertTrue(rs.files_directories.is_file(csv[0]))

        # clear temporary directory
        rs.close()



# Remotior Sensus , software to process remote sensing and GIS data.
# Copyright (C) 2022-2023 Luca Congedo.
# Author: Luca Congedo
# Email: ing.congedoluca@gmail.com
#
# This file is part of Remotior Sensus.
# Remotior Sensus is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License,
# or (at your option) any later version.
# Remotior Sensus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty
# of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with Remotior Sensus. If not, see <https://www.gnu.org/licenses/>.

import os
import random
import zipfile
from xml.dom import minidom
from xml.etree import cElementTree

import numpy as np

from remotior_sensus.core import (
    configurations as cfg, table_manager as tm
)
from remotior_sensus.core.bandset_catalog import BandSet
from remotior_sensus.core.processor_functions import (
    spectral_signature, get_values_for_scatter_plot
)
from remotior_sensus.util import (
    raster_vector, dates_times, files_directories, read_write_files,
    plot_tools, shared_tools
)

try:
    if cfg.gdal_path is not None:
        os.add_dll_directory(cfg.gdal_path)
except Exception as error:
    cfg.logger.log.error(str(error))
try:
    from osgeo import ogr
    from osgeo import osr
except Exception as error:
    cfg.logger.log.error(str(error))
try:
    from osgeo import gdal
except Exception as error:
    cfg.logger.log.error(str(error))


class SpectralSignaturesCatalog(object):
    """A class to manage Spectral Signatures and ROIs.

    """

    def __init__(
            self, bandset: BandSet = None, catalog_table=None,
            geometry_file_path=None, macroclass_field=None, class_field=None
    ):
        # relative BandSet
        self.bandset = bandset
        # spectral signatures catalog table (value, wavelength, and sd)
        self.table = catalog_table
        # dictionary of spectral signature tables (macroclass, class, selected)
        self.signatures = {}
        # dictionary of names of macroclasses
        self.macroclasses = {}
        # dictionary of color strings of macroclasses
        self.macroclasses_color_string = {}
        # vector file linked to signature table containing ROIs
        if geometry_file_path is None:
            geometry_file_path = cfg.temp.temporary_file_path(
                name_suffix=cfg.gpkg_suffix
            )
        if not macroclass_field:
            macroclass_field = cfg.macroclass_field_name
        if not class_field:
            class_field = cfg.class_field_name
        # default macroclass field name
        self.macroclass_field = macroclass_field
        # default class field name
        self.class_field = class_field
        self.geometry_file = geometry_file_path
        self.crs = None
        # create geometry vector
        if bandset:
            self.crs = bandset.crs
            if self.crs is not None:
                raster_vector.create_geometry_vector(
                    output_path=self.geometry_file, crs_wkt=self.crs,
                    macroclass_field_name=self.macroclass_field,
                    class_field_name=self.class_field
                )
            else:
                cfg.logger.log.debug('bandset without crs')

    # add spectral signature to Spectral Signatures Catalog
    def add_spectral_signature(
            self, value_list, macroclass_id=None, class_id=None,
            macroclass_name=None, class_name=None, wavelength_list=None,
            standard_deviation_list=None, signature_id=None, selected=1,
            min_dist_thr=0, max_like_thr=0, spec_angle_thr=0, geometry=0,
            signature=0, color_string=None, pixel_count=0, unit=None
    ):
        """Adds a spectral signature.

        This method adds spectral signature to Spectral Signatures Catalog.

        Args:
            value_list:
            macroclass_id:
            class_id:
            macroclass_name:
            class_name:
            wavelength_list:
            standard_deviation_list:
            signature_id:
            selected:
            min_dist_thr:
            max_like_thr:
            spec_angle_thr:
            geometry:
            signature:
            color_string:
            pixel_count: pixel count
            unit: unit

        Returns:
            object OutputManger

        """
        cfg.logger.log.debug('start')
        if macroclass_id is None:
            macroclass_id = 1
        if class_id is None:
            class_id = 1
        if color_string is None:
            color_string = shared_tools.random_color()
        # signature id
        if signature_id is None:
            signature_id = generate_signature_id()
        if wavelength_list is None:
            wavelength_list = self.bandset.get_wavelengths()
        # create signature table
        self.signatures[signature_id] = tm.create_spectral_signature_table(
            value_list=value_list, wavelength_list=wavelength_list,
            standard_deviation_list=standard_deviation_list
        )
        self.signature_to_catalog(
            signature_id=signature_id, macroclass_id=macroclass_id,
            class_id=class_id, macroclass_name=macroclass_name,
            class_name=class_name, selected=selected,
            min_dist_thr=min_dist_thr, max_like_thr=max_like_thr,
            spec_angle_thr=spec_angle_thr, geometry=geometry,
            signature=signature, color_string=color_string,
            pixel_count=pixel_count, unit=unit
        )
        cfg.logger.log.debug('end')

    # sets crs from bandset
    def set_crs_from_bandset(self, bandset):
        self.crs = bandset.crs

    # sets macroclass color string
    def set_macroclass_color(self, macroclass_id, color_string):
        self.macroclasses_color_string[macroclass_id] = color_string

    # add spectral signature reference to Spectral Signatures Catalog
    def signature_to_catalog(
            self, signature_id, macroclass_id, class_id, macroclass_name=None,
            class_name=None, selected=1, min_dist_thr=0, max_like_thr=0,
            spec_angle_thr=0, geometry=0, signature=0, color_string=None,
            pixel_count=0, unit=None
    ):
        # add signature to catalog
        self.table = tm.add_spectral_signature_to_catalog_table(
            signature_id=signature_id, macroclass_id=macroclass_id,
            class_id=class_id, class_name=class_name,
            previous_catalog=self.table, selected=selected,
            min_dist_thr=min_dist_thr, max_like_thr=max_like_thr,
            spec_angle_thr=spec_angle_thr, geometry=geometry,
            signature=signature, color_string=color_string,
            pixel_count=pixel_count, unit=unit
        )
        # add or update macroclass name
        if macroclass_name is not None:
            self.macroclasses[macroclass_id] = str(macroclass_name)
        # check macroclass name
        if macroclass_id not in self.macroclasses:
            self.macroclasses[macroclass_id] = '%s%s' % (
                cfg.macroclass_default, str(len(self.macroclasses) + 1))
        if macroclass_id not in self.macroclasses_color_string:
            self.macroclasses_color_string[macroclass_id] = color_string

    # calculate scatter plot for geometry from Spectral Signatures Catalog
    def calculate_scatter_plot_by_id(
            self, signature_id: str, band_x, band_y,
            decimal_round: int = None, plot=False, n_processes: int = None
    ):
        cfg.logger.log.debug('start')
        cfg.logger.log.debug(
            'signature_id: %s; decimal_round: %s'
            % (signature_id, decimal_round)
        )
        geometry = self.table[
            self.table['signature_id'] == signature_id].geometry[0]
        # geometry
        if geometry == 1:
            if not files_directories.is_file(self.geometry_file):
                cfg.logger.log.error(
                    'geometry file not found: %s' % self.geometry_file
                )
                raise Exception('geometry file not found')
            vector = raster_vector.get_polygon_from_vector(
                vector_path=self.geometry_file,
                attribute_filter="%s = '%s'" % (
                    cfg.uid_field_name, signature_id)
            )
            value_list = self.calculate_scatter_plot(
                roi_path=vector, band_x=band_x, band_y=band_y,
                n_processes=n_processes
            )
            if value_list is False:
                cfg.logger.log.error('unable to calculate')
                cfg.messages.error('unable to calculate')
                return False
            # calculate histogram
            histogram = shared_tools.calculate_2d_histogram(
                x_values=value_list[0], y_values=value_list[1],
                decimal_round=decimal_round
            )
            if plot is True:
                ax = plot_tools.prepare_scatter_plot()
                plots = plot_tools.add_values_to_scatter_plot(
                    histogram=histogram, ax=ax
                )
                macroclass_value = self.table[
                    self.table['signature_id'] == signature_id].macroclass_id[
                    0]
                macroclass_name = self.macroclasses[macroclass_value]
                class_value = self.table[
                    self.table['signature_id'] == signature_id].class_id[
                    0]
                class_name = self.table[
                    self.table['signature_id'] == signature_id].class_name[
                    0]
                plot_names = '%s#%s %s#%s' % (
                    macroclass_value, macroclass_name, class_value, class_name
                )
                plot_tools.create_scatter_plot(
                    ax=ax, plots=[plots], plot_names=[plot_names]
                )
            cfg.logger.log.debug('end')
            return histogram
        # not geometry
        else:
            cfg.logger.log.error('unable to calculate, missing geometry')
            cfg.messages.error('unable to calculate, missing geometry')
            return False

    # remove spectral signature and geometry from Spectral Signatures Catalog
    def remove_signature_by_id(self, signature_id: str):
        cfg.logger.log.debug('start')
        # remove signature
        try:
            del self.signatures[signature_id]
        except Exception as err:
            str(err)
        try:
            geometry = self.table[
                self.table['signature_id'] == signature_id].geometry[0]
            macroclass_value = self.table[
                self.table['signature_id'] == signature_id].macroclass_id[0]
        except Exception as err:
            str(err)
            cfg.logger.log.error('signature not found: %s' % signature_id)
            cfg.messages.error('signature not found: %s' % signature_id)
            return False
        # remove signature from table
        self.table = self.table[self.table['signature_id'] != signature_id]
        if macroclass_value not in self.table.macroclass_id.tolist():
            try:
                del self.macroclasses[macroclass_value]
                del self.macroclasses_color_string[macroclass_value]
            except Exception as err:
                str(err)
        if geometry == 1:
            if not files_directories.is_file(self.geometry_file):
                cfg.logger.log.error(
                    'geometry file not found: %s' % self.geometry_file
                )
                raise Exception('geometry file not found')
            # remove geometry
            raster_vector.remove_polygon_from_vector(
                vector_path=self.geometry_file,
                attribute_field=cfg.uid_field_name,
                attribute_value=signature_id
            )
        cfg.logger.log.debug('end')

    # merge spectral signatures and geometry from Spectral Signatures Catalog
    def merge_signatures_by_id(
            self, signature_id_list, calculate_signature=True,
            macroclass_id=None, class_id=None, macroclass_name=None,
            class_name=None, color_string=None
    ):
        cfg.logger.log.debug('start')
        geometry_check = True
        macroclass_value = 0
        class_value = 0
        geometry_ids = []
        signature_ids = []
        # get signatures and geometries from list
        count = 0
        for signature_id in signature_id_list:
            if cfg.action is False:
                break
            count += 1
            try:
                geometry = self.table[
                    self.table['signature_id'] == signature_id].geometry[0]
                # not geometry
                if geometry == 0:
                    geometry_check = False
                    signature_ids.append(signature_id)
                # geometry
                elif geometry == 1:
                    geometry_ids.append(signature_id)
                    signature_check = self.table[
                        self.table['signature_id'] == signature_id
                        ].signature[0]
                    # calculate signature
                    if calculate_signature is True and signature_check == 0:
                        if not files_directories.is_file(self.geometry_file):
                            cfg.logger.log.error(
                                'geometry file not found: %s'
                                % self.geometry_file
                            )
                            raise Exception('geometry file not found')
                        vector = raster_vector.get_polygon_from_vector(
                            vector_path=self.geometry_file,
                            attribute_filter="%s = '%s'" % (
                                cfg.uid_field_name, signature_id)
                        )
                        (value_list, standard_deviation_list, wavelength_list,
                         pixel_count) = self.calculate_signature(vector)
                        mc_value = self.table[
                            self.table['signature_id'] == signature_id
                            ].macroclass_id[0]
                        c_value = self.table[
                            self.table['signature_id'] == signature_id
                            ].class_id[0]
                        c_name = self.table[
                            self.table['signature_id'] == signature_id
                            ].class_name[0]
                        unit = self.table[
                            self.table['signature_id'] == signature_id
                            ].unit[0]
                        if color_string is None:
                            color_string = self.table[
                                self.table['signature_id'] == signature_id
                                ].color[0]
                        mc_name = self.macroclasses[mc_value]
                        self.add_spectral_signature(
                            value_list=value_list, macroclass_id=mc_value,
                            class_id=c_value, macroclass_name=mc_name,
                            class_name=c_name,
                            standard_deviation_list=standard_deviation_list,
                            signature_id=signature_id, geometry=1, signature=1,
                            color_string=color_string, pixel_count=pixel_count,
                            unit=unit
                        )
                    signature_ids.append(signature_id)
                # get first element class and macroclass
                if count == 1:
                    macroclass_value = self.table[
                        self.table['signature_id'] == signature_id
                        ].macroclass_id[0]
                    class_value = self.table[
                        self.table['signature_id'] == signature_id
                        ].class_id[0]
            except Exception as err:
                cfg.logger.log.error(str(err))
        if macroclass_id is not None:
            macroclass_value = macroclass_id
        if macroclass_name is None:
            macroclass_name = self.macroclasses[macroclass_value]
        if class_id is not None:
            class_value = class_id
        if class_name is None:
            class_name = 'merged'
        if color_string is None:
            color_string = shared_tools.random_color()
        # merge geometries if geometry == 1 for whole signature_id_list
        if geometry_check is True:
            temp_path = cfg.temp.temporary_file_path(
                name_suffix=cfg.gpkg_suffix
            )
            merged = raster_vector.merge_polygons(
                input_layer=self.geometry_file, value_list=signature_id_list,
                target_layer=temp_path
            )
            # import vector
            self.import_vector(
                file_path=merged, macroclass_value=macroclass_value,
                class_value=class_value, macroclass_name=macroclass_name,
                class_name=class_name, calculate_signature=calculate_signature,
                color_string=color_string
            )
        # merge signatures if not geometry
        else:
            wavelength = None
            value_arrays = []
            std_arrays = []
            for signature in signature_ids:
                if cfg.action is False:
                    break
                value_arrays.append(self.signatures[signature].value)
                std_arrays.append(
                    self.signatures[signature].standard_deviation
                )
                wavelength = self.signatures[signature].wavelength
            unit = self.table[
                self.table['signature_id'] == signature_ids[0]].unit[0]
            wavelength_list = wavelength.tolist()
            values = np.column_stack(value_arrays)
            stds = np.column_stack(std_arrays)
            values_mean = np.mean(values, axis=1)
            stds_squared = np.square(stds)
            stds_squared_sum = np.sum(stds_squared, axis=1)
            stds_variance = np.divide(stds_squared_sum, stds_squared.shape[1])
            stds_mean = np.sqrt(stds_variance)
            self.add_spectral_signature(
                value_list=values_mean.tolist(),
                macroclass_id=macroclass_value, class_id=class_value,
                macroclass_name=macroclass_name, class_name='merged',
                wavelength_list=wavelength_list,
                standard_deviation_list=stds_mean.tolist(), geometry=0,
                signature=1, color_string=color_string, pixel_count=0,
                unit=unit
            )
        cfg.logger.log.debug('end')

    # export signatures as csv
    def export_signatures_as_csv(
            self, signature_id_list, output_directory, separator=','
    ):
        cfg.logger.log.debug(
            'export_signatures_as_csv: %s' % str(signature_id_list)
        )
        files_directories.create_directory(output_directory)
        output_list = []
        for signature_id in signature_id_list:
            if cfg.action is False:
                break
            try:
                values = self.signatures[signature_id].value.tolist()
                wavelength = self.signatures[signature_id].wavelength.tolist()
                standard_deviation = (
                    self.signatures[signature_id].standard_deviation.tolist()
                )
                macroclass_id = self.table[
                    self.table['signature_id'] == signature_id
                    ].macroclass_id[0]
                macroclass_name = self.macroclasses[macroclass_id]
                class_id = self.table[
                    self.table['signature_id'] == signature_id
                    ].class_id[0]
                class_name = self.table[
                    self.table['signature_id'] == signature_id
                    ].class_name[0]
                output_file = '%s/%s_%s_%s_%s.csv' % (
                    output_directory, macroclass_id, macroclass_name,
                    class_id, class_name
                )
                text = ''
                for v in range(len(values)):
                    if cfg.action is False:
                        break
                    text += '%s%s%s%s%s\n' % (
                        values[v], separator, wavelength[v], separator,
                        standard_deviation[v]
                    )
                read_write_files.write_file(text, output_file)
                output_list.append(output_file)
            except Exception as err:
                cfg.logger.log.error(str(err))
        return output_list

    # import spectral signature csv to Spectral Signatures Catalog
    def import_spectral_signature_csv(
            self, csv_path, macroclass_id=None, class_id=None,
            macroclass_name=None, class_name=None, separator=',',
            color_string=None
    ):
        cfg.logger.log.debug('start')
        # import csv as comma separated with fields value, wavelength,
        # standard_deviation (optional)
        if files_directories.is_file(csv_path):
            csv = tm.open_file(
                file_path=csv_path, separators=separator,
                field_names=['value', 'wavelength',
                             'standard_deviation'],
                progress_message=False, skip_first_line=False
            )
            bandset_wavelength = self.bandset.bands['wavelength']
            unit = self.bandset.get_wavelength_units()[0]
            value_list = []
            wavelength_list = []
            standard_deviation_list = []
            if 'wavelength' in tm.columns(csv):
                for b in bandset_wavelength.tolist():
                    if cfg.action is False:
                        break
                    arg_min = np.abs(csv.wavelength - b).argmin()
                    wavelength_list.append(b)
                    value_list.append(csv.value[arg_min])
                    if 'standard_deviation' in tm.columns(csv):
                        standard_deviation_list.append(
                            csv.standard_deviation[arg_min]
                        )
            if len(value_list) == 0:
                cfg.logger.log.error('file: %s' % csv_path)
                cfg.messages.error('error importing file %s' % csv_path)
                return
            if len(wavelength_list) == 0:
                wavelength_list = None
            if len(standard_deviation_list) == 0:
                standard_deviation_list = None
            if color_string is None:
                color_string = shared_tools.random_color()
            self.add_spectral_signature(
                value_list=value_list, macroclass_id=macroclass_id,
                class_id=class_id, macroclass_name=macroclass_name,
                class_name=class_name, wavelength_list=wavelength_list,
                standard_deviation_list=standard_deviation_list, geometry=0,
                signature=1, color_string=color_string, pixel_count=0,
                unit=unit
            )
            cfg.logger.log.debug('end; imported: %s' % csv_path)
        else:
            cfg.logger.log.error('error file not found: %s' % csv_path)
            cfg.messages.error('error file not found: %s' % csv_path)

    # import Spectral Signatures Catalog file
    def import_file(self, file_path):
        cfg.logger.log.debug(
            'import_file: %s' % file_path
        )
        # create temporary directory
        temp_dir = cfg.temp.create_temporary_directory()
        file_list = files_directories.unzip_file(file_path, temp_dir)
        # list of new ids
        signature_ids = {}
        geometry_ids = {}
        geometry_file = None
        table = None
        for f in file_list:
            if cfg.action is True:
                f_name = files_directories.file_name(f, suffix=True)
                if f_name == 'geometry.gpkg':
                    geometry_file = f
                elif f_name == 'table':
                    table = np.core.records.fromfile(
                        f, dtype=cfg.spectral_dtype_list
                    )
                    # remove file
                    files_directories.remove_file(f)
                elif f_name == 'macroclasses.xml':
                    tree = cElementTree.parse(f)
                    root = tree.getroot()
                    version = root.get('version')
                    if version is None:
                        cfg.logger.log.error(
                            'failed loading signatures: %s'
                            % file_path
                        )
                        cfg.messages.error(
                            'failed loading signatures: %s'
                            % file_path
                        )
                    else:
                        for child in root:
                            if cfg.action is False:
                                break
                            macroclass_id = child.get('id')
                            macroclass_name = child.get('name')
                            macroclass_color = child.get('color')
                            if int(macroclass_id) not in self.macroclasses:
                                self.macroclasses[int(macroclass_id)] = str(
                                    macroclass_name
                                )
                            if (int(macroclass_id)
                                    not in self.macroclasses_color_string):
                                self.macroclasses_color_string[
                                    int(macroclass_id)] = str(macroclass_color)
                    # remove file
                    files_directories.remove_file(f)
                else:
                    signature_id = generate_signature_id()
                    signature_ids[f_name] = signature_id
                    f_name = signature_id
                    self.signatures[f_name] = np.core.records.fromfile(
                        f, dtype=cfg.signature_dtype_list
                    )
                    # remove file
                    files_directories.remove_file(f)
            else:
                cfg.logger.log.error('cancel')
                return
        cfg.logger.log.debug('signature_ids: %s' % signature_ids)
        # import vector
        if geometry_file is not None:
            # get vector crs
            vector_crs = raster_vector.get_crs(geometry_file)
            if self.crs is None:
                cfg.logger.log.error('crs not defined')
                raise Exception('crs not defined')
            # check crs
            catalog_sr = osr.SpatialReference()
            catalog_sr.ImportFromWkt(self.crs)
            vector_sr = osr.SpatialReference()
            vector_sr.ImportFromWkt(vector_crs)
            # required by GDAL 3 coordinate order
            try:
                catalog_sr.SetAxisMappingStrategy(
                    osr.OAMS_TRADITIONAL_GIS_ORDER
                )
                vector_sr.SetAxisMappingStrategy(
                    osr.OAMS_TRADITIONAL_GIS_ORDER
                )
            except Exception as err:
                str(err)
            if catalog_sr.IsSame(vector_sr) == 1:
                coord_transform = None
            else:
                # coordinate transformation
                coord_transform = osr.CoordinateTransformation(
                    vector_sr, catalog_sr
                )
            # open input vector
            i_vector = ogr.Open(geometry_file)
            i_layer = i_vector.GetLayer()
            catalog_vector = ogr.Open(self.geometry_file, 1)
            catalog_layer = catalog_vector.GetLayer()
            catalog_layer_definition = catalog_layer.GetLayerDefn()
            # import geometries
            i_feature = i_layer.GetNextFeature()
            while i_feature:
                if cfg.action is True:
                    # get geometry
                    geom = i_feature.GetGeometryRef()
                    if coord_transform is not None:
                        # project feature
                        geom.Transform(coord_transform)
                    o_feature = ogr.Feature(catalog_layer_definition)
                    o_feature.SetGeometry(geom)
                    sig_id = i_feature.GetField(cfg.uid_field_name)
                    mc_value = i_feature.GetField(self.macroclass_field)
                    c_value = i_feature.GetField(self.class_field)
                    if sig_id in signature_ids:
                        o_feature.SetField(
                            cfg.uid_field_name, signature_ids[sig_id]
                        )
                    else:
                        signature_id = generate_signature_id()
                        geometry_ids[sig_id] = signature_id
                        o_feature.SetField(
                            cfg.uid_field_name, signature_id
                        )
                    o_feature.SetField(cfg.class_field_name, int(c_value))
                    o_feature.SetField(
                        cfg.macroclass_field_name, int(mc_value)
                    )
                    catalog_layer.CreateFeature(o_feature)
                    o_feature.Destroy()
                    i_feature.Destroy()
                    i_feature = i_layer.GetNextFeature()
                else:
                    # close files
                    i_vector.Destroy()
                    catalog_vector.Destroy()
                    i_feature = False
                    cfg.logger.log.error('cancel')
        # import table
        if table is not None:
            for sig_id in signature_ids:
                table['signature_id'][
                    table['signature_id'] == sig_id] = signature_ids[sig_id]
            for sig_id in geometry_ids:
                table['signature_id'][
                    table['signature_id'] == sig_id] = geometry_ids[sig_id]
            self.table = tm.append_tables(self.table, table)

    # import vector to Spectral Signatures Catalog
    def import_vector(
            self, file_path, macroclass_value=None, class_value=None,
            macroclass_name=None, class_name=None, macroclass_field=None,
            class_field=None, macroclass_name_field=None,
            class_name_field=None, calculate_signature=True,
            color_string=None
    ):
        cfg.logger.log.debug('start')
        if files_directories.is_file(file_path):
            # check geometry vector
            if not files_directories.is_file(self.geometry_file):
                if self.bandset is None:
                    cfg.logger.log.error('bandset not found')
                    raise Exception('bandset not found')
                if self.crs is not None:
                    raster_vector.create_geometry_vector(
                        output_path=self.geometry_file, crs_wkt=self.crs,
                        macroclass_field_name=self.macroclass_field,
                        class_field_name=self.class_field
                    )
                else:
                    cfg.logger.log.error('crs not defined')
                    raise Exception('crs not defined')
            # get vector crs
            vector_crs = raster_vector.get_crs(file_path)
            if self.crs is None:
                cfg.logger.log.error('crs not defined')
                raise Exception('crs not defined')
            # check crs
            catalog_sr = osr.SpatialReference()
            catalog_sr.ImportFromWkt(self.crs)
            vector_sr = osr.SpatialReference()
            vector_sr.ImportFromWkt(vector_crs)
            unit = self.bandset.get_wavelength_units()[0]
            # required by GDAL 3 coordinate order
            try:
                catalog_sr.SetAxisMappingStrategy(
                    osr.OAMS_TRADITIONAL_GIS_ORDER
                )
                vector_sr.SetAxisMappingStrategy(
                    osr.OAMS_TRADITIONAL_GIS_ORDER
                )
            except Exception as err:
                str(err)
            if catalog_sr.IsSame(vector_sr) == 1:
                coord_transform = None
            else:
                # coordinate transformation
                coord_transform = osr.CoordinateTransformation(
                    vector_sr, catalog_sr
                )
            # open input vector
            i_vector = ogr.Open(file_path)
            i_layer = i_vector.GetLayer()
            catalog_vector = ogr.Open(self.geometry_file, 1)
            catalog_layer = catalog_vector.GetLayer()
            catalog_layer_definition = catalog_layer.GetLayerDefn()
            # import geometries
            i_feature = i_layer.GetNextFeature()
            while i_feature:
                if cfg.action is True:
                    signature_id = generate_signature_id()
                    # get geometry
                    geom = i_feature.GetGeometryRef()
                    if coord_transform is not None:
                        # project feature
                        geom.Transform(coord_transform)
                    o_feature = ogr.Feature(catalog_layer_definition)
                    o_feature.SetGeometry(geom)
                    if macroclass_value is None:
                        mc_value = i_feature.GetField(macroclass_field)
                    else:
                        mc_value = macroclass_value
                    if class_value is None:
                        c_value = i_feature.GetField(class_field)
                    else:
                        c_value = class_value
                    if macroclass_name is None:
                        mc_name = i_feature.GetField(macroclass_name_field)
                    else:
                        mc_name = macroclass_name
                    if class_name is None:
                        c_name = i_feature.GetField(class_name_field)
                    else:
                        c_name = class_name
                    o_feature.SetField(cfg.uid_field_name, signature_id)
                    o_feature.SetField(cfg.class_field_name, int(c_value))
                    o_feature.SetField(
                        cfg.macroclass_field_name, int(mc_value)
                    )
                    catalog_layer.CreateFeature(o_feature)
                    if color_string is None:
                        color_string = shared_tools.random_color()
                    if calculate_signature:
                        temp_path = cfg.temp.temporary_file_path(
                            name_suffix=cfg.gpkg_suffix
                        )
                        raster_vector.create_geometry_vector(
                            output_path=temp_path, crs_wkt=self.crs,
                            macroclass_field_name=self.macroclass_field,
                            class_field_name=self.class_field
                        )
                        temp_vector = ogr.Open(temp_path, 1)
                        temp_layer = temp_vector.GetLayer()
                        temp_layer.CreateFeature(o_feature)
                        temp_vector.Destroy()
                        try:
                            (value_list, standard_deviation_list,
                             wavelength_list,
                             pixel_count) = self.calculate_signature(temp_path)
                        except Exception as err:
                            cfg.logger.log.error(str(err))
                            self.signature_to_catalog(
                                signature_id=signature_id,
                                macroclass_id=mc_value,
                                class_id=c_value, macroclass_name=mc_name,
                                class_name=c_name, geometry=1, signature=0,
                                color_string=color_string, unit=unit
                            )
                            return False
                        self.add_spectral_signature(
                            value_list=value_list, macroclass_id=mc_value,
                            class_id=c_value, macroclass_name=mc_name,
                            class_name=c_name,
                            standard_deviation_list=standard_deviation_list,
                            signature_id=signature_id, geometry=1, signature=1,
                            color_string=color_string, pixel_count=pixel_count,
                            unit=unit
                        )
                    else:
                        self.signature_to_catalog(
                            signature_id=signature_id, macroclass_id=mc_value,
                            class_id=c_value, macroclass_name=mc_name,
                            class_name=c_name, geometry=1, signature=0,
                            color_string=color_string, unit=unit
                        )
                    o_feature.Destroy()
                    i_feature.Destroy()
                    i_feature = i_layer.GetNextFeature()
                else:
                    # close files
                    i_vector.Destroy()
                    catalog_vector.Destroy()
                    i_feature = False
                    cfg.logger.log.error('cancel')
            # close files
            i_vector.Destroy()
            catalog_vector.Destroy()
            cfg.logger.log.debug('end; imported: %s' % file_path)
        else:
            cfg.logger.log.error('error file not found: %s' % file_path)
            cfg.messages.error('error file not found: %s' % file_path)

    # calculate spectral signatures
    def calculate_signature(self, roi_path, n_processes: int = None):
        cfg.logger.log.debug('calculate_signature: %s' % roi_path)
        if n_processes is None:
            n_processes = cfg.n_processes
        min_x, max_x, min_y, max_y = raster_vector.get_layer_extent(roi_path)
        path_list = self.bandset.get_absolute_paths()
        virtual_path_list = []
        for p in path_list:
            if cfg.action is False:
                break
            temp_path = cfg.temp.temporary_file_path(
                name_suffix=cfg.tif_suffix
            )
            virtual = raster_vector.create_virtual_raster(
                input_raster_list=[p], output=temp_path,
                box_coordinate_list=[min_x, max_y, max_x, min_y]
            )
            virtual_path_list.append(virtual)
        roi_paths = [roi_path] * len(path_list)
        cfg.multiprocess.run_separated(
            raster_path_list=virtual_path_list, function=spectral_signature,
            function_argument=roi_paths, function_variable=virtual_path_list,
            n_processes=n_processes, keep_output_argument=True,
            progress_message='calculate signature', min_progress=1,
            max_progress=80
        )
        wavelength_list = self.bandset.get_wavelengths()
        cfg.multiprocess.multiprocess_spectral_signature()
        if cfg.multiprocess.output is False:
            return False
        else:
            (value_list, standard_deviation_list,
             count_list) = cfg.multiprocess.output
            return (
                value_list, standard_deviation_list, wavelength_list,
                count_list
            )

    # import vector for scatter plot
    def calculate_scatter_plot(
            self, roi_path, band_x, band_y, n_processes: int = None
    ):
        cfg.logger.log.debug(
            'calculate_scatter_plot: %s; band x: %s; band y: %s'
            % (roi_path, band_x, band_y)
        )
        if n_processes is None:
            n_processes = cfg.n_processes
        min_x, max_x, min_y, max_y = raster_vector.get_layer_extent(roi_path)
        band_x_path = self.bandset.get_absolute_path(band_number=band_x)
        band_y_path = self.bandset.get_absolute_path(band_number=band_y)
        if band_x_path is None or band_y_path is None:
            cfg.logger.log.error('failed to get bands')
            cfg.messages.error('failed to get bands')
            return False
        path_list = [band_x_path, band_y_path]
        virtual_path_list = []
        for p in path_list:
            if cfg.action is False:
                break
            temp_path = cfg.temp.temporary_file_path(
                name_suffix=cfg.tif_suffix
            )
            virtual = raster_vector.create_virtual_raster(
                input_raster_list=[p], output=temp_path,
                box_coordinate_list=[min_x, max_y, max_x, min_y]
            )
            virtual_path_list.append(virtual)
        roi_paths = [roi_path] * len(path_list)
        if n_processes > 2:
            n_processes = 2
        cfg.multiprocess.run_separated(
            raster_path_list=virtual_path_list,
            function=get_values_for_scatter_plot,
            function_argument=roi_paths, function_variable=virtual_path_list,
            n_processes=n_processes, keep_output_argument=True,
            progress_message='calculate band values', min_progress=1,
            max_progress=80
        )
        cfg.multiprocess.multiprocess_scatter_values()
        if cfg.multiprocess.output is False:
            return False
        value_list = cfg.multiprocess.output
        return value_list

    # save Spectral Signatures Catalog to file
    def save(self, output_path, signature_id_list=None):
        cfg.logger.log.debug(
            'save Spectral Signatures Catalog: %s' % output_path
        )
        # file list
        file_list = []
        # create temporary directory
        temp_dir = cfg.temp.create_temporary_directory()
        # geometry file
        if files_directories.is_file(self.geometry_file):
            if signature_id_list is None:
                files_directories.copy_file(
                    self.geometry_file, '%s/geometry.gpkg' % temp_dir
                )
            else:
                self.export_vector(
                    signature_id_list, '%s/geometry.gpkg' % temp_dir
                )
            file_list.append('%s/geometry.gpkg' % temp_dir)
        # create xml file
        root = cElementTree.Element('macroclasses')
        root.set('version', str(cfg.version))
        root.set('macroclass_field', str(self.macroclass_field))
        root.set('class_field', str(self.class_field))
        if self.signatures is not None:
            for signature_id in self.signatures:
                if cfg.action is False:
                    break
                if (signature_id_list is None
                        or signature_id in signature_id_list):
                    # create file inside temporary directory
                    self.signatures[signature_id].tofile(
                        file='%s/%s' % (temp_dir, signature_id)
                    )
                    file_list.append('%s/%s' % (temp_dir, signature_id))
        if signature_id_list is None:
            macroclass_list = []
        else:
            macroclass_list = self.table[
                np.in1d(self.table['signature_id'], signature_id_list)
            ].macroclass_id.tolist()
        if self.macroclasses is not None:
            for macroclass in self.macroclasses:
                if signature_id_list is None or macroclass in macroclass_list:
                    macroclass_element = cElementTree.SubElement(
                        root, 'macroclass'
                    )
                    macroclass_element.set('id', str(macroclass))
                    macroclass_element.set(
                        'name',
                        str(self.macroclasses[macroclass])
                    )
                    macroclass_element.set(
                        'color',
                        str(self.macroclasses_color_string[macroclass])
                    )
        # save to file
        pretty_xml = minidom.parseString(
            cElementTree.tostring(root)
        ).toprettyxml()
        # create file inside temporary directory
        read_write_files.write_file(
            pretty_xml,
            '%s/macroclasses.xml' % temp_dir
        )
        file_list.append('%s/macroclasses.xml' % temp_dir)
        # create file inside temporary directory
        if self.table is not None:
            if signature_id_list is None:
                self.table.tofile(file='%s/table' % temp_dir)
            else:
                self.table[
                    np.in1d(self.table['signature_id'], signature_id_list)
                ].tofile(file='%s/table' % temp_dir)
            file_list.append('%s/table' % temp_dir)
        # zip files
        files_directories.zip_files(
            file_list, output_path,
            compression=zipfile.ZIP_STORED
        )
        cfg.logger.log.debug('saved signature catalog')
        return output_path

    # load Spectral Signatures Catalog from file
    def load(self, file_path):
        cfg.logger.log.debug(
            'load Spectral Signatures Catalog: %s' % file_path
        )
        self.table = None
        self.signatures = {}
        # create temporary directory
        temp_dir = cfg.temp.create_temporary_directory()
        file_list = files_directories.unzip_file(file_path, temp_dir)
        for f in file_list:
            if cfg.action is False:
                break
            f_name = files_directories.file_name(f, suffix=True)
            if f_name == 'geometry.gpkg':
                self.geometry_file = f
            elif f_name == 'table':
                self.table = np.core.records.fromfile(
                    f, dtype=cfg.spectral_dtype_list
                )
                # remove file
                files_directories.remove_file(f)
            elif f_name == 'macroclasses.xml':
                tree = cElementTree.parse(f)
                root = tree.getroot()
                version = root.get('version')
                if version is None:
                    cfg.logger.log.error(
                        'failed loading signatures: %s'
                        % file_path
                    )
                    cfg.messages.error(
                        'failed loading signatures: %s'
                        % file_path
                    )
                else:
                    self.macroclass_field = root.get('macroclass_field')
                    self.class_field = root.get('class_field')
                    self.macroclasses = {}
                    self.macroclasses_color_string = {}
                    for child in root:
                        if cfg.action is False:
                            break
                        macroclass_id = child.get('id')
                        macroclass_name = child.get('name')
                        macroclass_color = child.get('color')
                        self.macroclasses[int(macroclass_id)] = str(
                            macroclass_name
                        )
                        self.macroclasses_color_string[
                            int(macroclass_id)] = str(macroclass_color)
                # remove file
                files_directories.remove_file(f)
            else:
                self.signatures[f_name] = np.core.records.fromfile(
                    f, dtype=cfg.signature_dtype_list
                )
                # remove file
                files_directories.remove_file(f)

    # prepare signature values for plot
    def export_signature_values_for_plot(
            self, signature_id, plot_catalog=None
    ):
        cfg.logger.log.debug(
            'export_signature_values_for_plot: %s'
            % signature_id
        )
        # check signature
        try:
            signature = self.table[
                self.table['signature_id'] == signature_id].signature[0]
        except Exception as err:
            str(err)
            cfg.logger.log.error('signature not found: %s' % signature_id)
            cfg.messages.error('signature not found: %s' % signature_id)
            return False
        if signature == 0:
            geometry = self.table[
                self.table['signature_id'] == signature_id].geometry[0]
            # not geometry
            if geometry == 0:
                cfg.logger.log.error('signature not found: %s' % signature_id)
                cfg.messages.error('signature not found: %s' % signature_id)
                return False
            # geometry
            else:
                if files_directories.is_file(self.geometry_file):
                    # calculate signature
                    vector = raster_vector.get_polygon_from_vector(
                        vector_path=self.geometry_file,
                        attribute_filter="%s = '%s'" % (
                            cfg.uid_field_name, signature_id
                        )
                    )
                    try:
                        (value_list, standard_deviation_list, wavelength_list,
                         pixel_count) = self.calculate_signature(vector)
                    except Exception as err:
                        cfg.logger.log.error(str(err))
                        return False
                else:
                    cfg.logger.log.error(
                        'geometry file not found: %s'
                        % self.geometry_file
                    )
                    raise Exception('geometry file not found')
        else:
            value_list = self.signatures[signature_id].value
            wavelength_list = self.signatures[signature_id].wavelength
            standard_deviation_list = self.signatures[
                signature_id].standard_deviation
            pixel_count = self.table[
                self.table['signature_id'] == signature_id].pixel_count[0]
        mc_value = self.table[
            self.table['signature_id'] == signature_id].macroclass_id[0]
        c_value = self.table[
            self.table['signature_id'] == signature_id].class_id[0]
        c_name = self.table[
            self.table['signature_id'] == signature_id].class_name[0]
        color_string = self.table[
            self.table['signature_id'] == signature_id].color[0]
        mc_name = self.macroclasses[mc_value]
        signature_plot = SpectralSignaturePlot(
            value=value_list, wavelength=wavelength_list,
            standard_deviation=standard_deviation_list,
            pixel_count=pixel_count, signature_id=signature_id,
            macroclass_id=mc_value, class_id=c_value, macroclass_name=mc_name,
            class_name=c_name, color_string=color_string
        )
        if plot_catalog is not None:
            plot_catalog.add_signature(signature_plot)
        return signature_plot

    # export geometries to vector
    def export_vector(
            self, signature_id_list, output_path, vector_format=None
    ):
        cfg.logger.log.debug(
            'export_vector: %s' % str(signature_id_list)
        )
        if vector_format is None:
            vector_format = 'GPKG'
        raster_vector.save_polygons(
            input_layer=self.geometry_file, value_list=signature_id_list,
            target_layer=output_path, vector_format=vector_format
        )
        return output_path

    # display plot of signatures using Matplotlib
    def add_signatures_to_plot_by_id(self, signature_id_list):
        cfg.logger.log.debug(
            'add_signatures_to_plot_by_id: %s'
            % str(signature_id_list)
        )
        try:
            plot_catalog = SpectralSignaturePlotCatalog()
            ax = plot_tools.prepare_plot()
            for signature in signature_id_list:
                if cfg.action is False:
                    break
                self.export_signature_values_for_plot(
                    signature_id=signature, plot_catalog=plot_catalog
                )
            name_list = plot_catalog.get_signature_names()
            value_list = plot_catalog.get_signature_values()
            wavelength_list = plot_catalog.get_signature_wavelength()
            color_list = plot_catalog.get_signature_color()
            plots, plot_names, x_ticks, y_ticks, v_lines = (
                plot_tools.add_lines_to_plot(
                    name_list=name_list, wavelength_list=wavelength_list,
                    value_list=value_list, color_list=color_list
                )
            )
            plot_tools.create_plot(
                ax=ax, plots=plots, plot_names=plot_names, x_ticks=x_ticks,
                y_ticks=y_ticks, v_lines=v_lines
            )
        except Exception as err:
            cfg.logger.log.error(str(err))
            return False

    # calculate Bray-Curtis similarity
    # (100 - 100 * sum(abs(x[ki]-x[kj]) / (sum(x[ki] + x[kj])))
    def calculate_bray_curtis_similarity(self, signature_id_x, signature_id_y):
        cfg.logger.log.debug(
            'calculate_bray_curtis_similarity i: %s; j: %s'
            % (str(signature_id_x), str(signature_id_y))
        )
        value = shared_tools.calculate_bray_curtis_similarity(
            values_x=self.signatures[signature_id_x].value,
            values_y=self.signatures[signature_id_y].value
        )
        return value

    # calculate Euclidean distance sqrt(sum((x[ki] - x[kj])^2))
    def calculate_euclidean_distance(self, signature_id_x, signature_id_y):
        cfg.logger.log.debug(
            'calculate_euclidean_distance i: %s; j: %s'
            % (str(signature_id_x), str(signature_id_y))
        )
        value = shared_tools.calculate_euclidean_distance(
            values_x=self.signatures[signature_id_x].value,
            values_y=self.signatures[signature_id_y].value
        )
        return value

    # calculate Spectral angle
    # [ arccos( sum(r_i * s_i) / sqrt( sum(r_i**2) * sum(s_i**2) ) ) ]
    def calculate_spectral_angle(self, signature_id_x, signature_id_y):
        cfg.logger.log.debug(
            'Spectral angle i: %s; j: %s'
            % (str(signature_id_x), str(signature_id_y))
        )
        value = shared_tools.calculate_spectral_angle(
            values_x=self.signatures[signature_id_x].value,
            values_y=self.signatures[signature_id_y].value
        )
        return value


class SpectralSignaturePlot(object):
    """A class to manage Spectral Signatures for plots.

    """

    def __init__(
            self, value, wavelength, standard_deviation=None, pixel_count=None,
            signature_id=None, macroclass_id=None, class_id=None,
            macroclass_name=None, class_name=None, color_string=None,
            selected=None
    ):
        self.value = value
        self.wavelength = wavelength
        self.standard_deviation = standard_deviation
        if pixel_count is None:
            pixel_count = 0
        self.pixel_count = pixel_count
        if signature_id is None:
            signature_id = generate_signature_id()
        self.signature_id = signature_id
        if macroclass_id is None:
            macroclass_id = 0
        self.macroclass_id = macroclass_id
        if class_id is None:
            class_id = 0
        self.class_id = class_id
        if macroclass_name is None:
            macroclass_name = cfg.macroclass_default
        self.macroclass_name = macroclass_name
        if class_name is None:
            class_name = cfg.class_default
        self.class_name = class_name
        if color_string is None:
            color_string = '#000000'
        self.color = color_string
        if selected is None:
            selected = 1
        self.selected = selected
        # generic attributes
        self.attributes = {}


class SpectralSignaturePlotCatalog(object):
    """A class to manage Spectral Signatures Catalog for plots.

    """

    def __init__(
            self, signature: SpectralSignaturePlot = None
    ):
        self.catalog = {}
        if signature is not None:
            self.catalog[signature.signature_id] = signature

    # add signature to catalog
    def add_signature(self, signature: SpectralSignaturePlot):
        self.catalog[signature.signature_id] = signature
        cfg.logger.log.debug('add_signature: %s' % signature.signature_id)
        return True

    # add signature to catalog
    def remove_signature(self, signature_id: str):
        try:
            del self.catalog[signature_id]
            cfg.logger.log.debug('remove_signature: %s' % signature_id)
            return True
        except Exception as err:
            str(err)
            cfg.logger.log.error('signature not found: %s' % signature_id)
            cfg.messages.error('signature not found: %s' % signature_id)
            return False

    def get_signature_count(self) -> int:
        """Gets count of signatures in the catalog.

        This function gets the count of signatures present in the catalog.

        Returns:
            The integer number of signatures.

        Examples:
            Count of signatures present.
                >>> catalog = SpectralSignaturePlotCatalog()
                >>> count = catalog.get_signature_count()
                >>> print(count)
                1
        """
        return len(self.catalog)

    def get_signature(self, signature_id) -> SpectralSignaturePlot:
        """Gets signature in the catalog.

        This function gets signature by id from the catalog.

        Returns:
            The SpectralSignaturePlot.

        Examples:
            Get signature.
                >>> catalog = SpectralSignaturePlotCatalog()
                >>> signature = catalog.get_signature(signature_id='signature_id')
                >>> print(signature.signature_id)
                'signature_id'
        """  # noqa: E501
        return self.catalog[signature_id]

    def get_signature_ids(self) -> list:
        """Gets signature ids in the catalog.

        This function gets signature ids from the catalog.

        Returns:
            The list of ids.

        Examples:
            Get signature.
                >>> catalog = SpectralSignaturePlotCatalog()
                >>> signature_ids = catalog.get_signature_ids()
                >>> print(signature_ids)
                ['signature_id']
        """
        return list(self.catalog.keys())

    def get_signature_names(self, selected=True) -> list:
        """Gets signature names in the catalog.

        This function gets signature names from the catalog,
        derived from SpectralSignaturePlot.

        Returns:
            The list of values.

        Examples:
            Get signature.
                >>> catalog = SpectralSignaturePlotCatalog()
                >>> signature_names = catalog.get_signature_names()
                >>> print(signature_names)
                ['name']
        """
        property_list = []
        for signature in self.catalog:
            if cfg.action is False:
                break
            if selected is True:
                if self.catalog[signature].selected == 1:
                    property_list.append(
                        '%s#%s %s#%s' % (
                            self.catalog[signature].macroclass_id,
                            self.catalog[signature].macroclass_name,
                            self.catalog[signature].class_id,
                            self.catalog[signature].class_name)
                    )
            else:
                property_list.append(
                    '%s#%s %s#%s' % (
                        self.catalog[signature].macroclass_id,
                        self.catalog[signature].macroclass_name,
                        self.catalog[signature].class_id,
                        self.catalog[signature].class_name)
                )
        return property_list

    def get_signature_values(self, selected=True) -> list:
        """Gets signature values in the catalog.

        This function gets signature values from the catalog.

        Returns:
            The list of values.

        """
        property_list = []
        for signature in self.catalog:
            if cfg.action is False:
                break
            if selected is True:
                if self.catalog[signature].selected == 1:
                    property_list.append(
                        self.catalog[signature].value
                    )
            else:
                property_list.append(self.catalog[signature].value)
        return property_list

    def get_signature_wavelength(self, selected=True) -> list:
        """Gets signature wavelength in the catalog.

        This function gets signature wavelength from the catalog.

        Returns:
            The list of values.

        """
        property_list = []
        for signature in self.catalog:
            if cfg.action is False:
                break
            if selected is True:
                if self.catalog[signature].selected == 1:
                    property_list.append(
                        self.catalog[signature].wavelength
                    )
            else:
                property_list.append(self.catalog[signature].wavelength)
        return property_list

    def get_signature_standard_deviation(self, selected=True) -> list:
        """Gets signature standard deviation in the catalog.

        This function gets signature standard deviation from the catalog.

        Returns:
            The list of values.

        """
        property_list = []
        for signature in self.catalog:
            if cfg.action is False:
                break
            if selected is True:
                if self.catalog[signature].selected == 1:
                    property_list.append(
                        self.catalog[signature].standard_deviation
                    )
            else:
                property_list.append(
                    self.catalog[signature].standard_deviation
                )
        return property_list

    def get_signature_color(self, selected=True) -> list:
        """Gets signature color in the catalog.

        This function gets signature color from the catalog.

        Returns:
            The list of values.

        """
        property_list = []
        for signature in self.catalog:
            if cfg.action is False:
                break
            if selected is True:
                if self.catalog[signature].selected == 1:
                    property_list.append(
                        self.catalog[signature].color
                    )
            else:
                property_list.append(
                    self.catalog[signature].color
                )
        return property_list

    def get_generic_attributes(self, selected=True) -> list:
        """Gets generic attributes in the catalog.

        This function gets generic attributes from the catalog.

        Returns:
            The list of values.

        """
        property_list = []
        for signature in self.catalog:
            if cfg.action is False:
                break
            if selected is True:
                if self.catalog[signature].selected == 1:
                    property_list.append(
                        self.catalog[signature].attributes
                    )
            else:
                property_list.append(
                    self.catalog[signature].attributes
                )
        return property_list


# generate signature id
def generate_signature_id():
    times = dates_times.get_time_string()
    r = str(random.randint(0, 1000))
    uid = 's%s_%s' % (times, r)
    return uid



# Remotior Sensus , software to process remote sensing and GIS data.
# Copyright (C) 2022-2023 Luca Congedo.
# Author: Luca Congedo
# Email: ing.congedoluca@gmail.com
#
# This file is part of Remotior Sensus.
# Remotior Sensus is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by 
# the Free Software Foundation, either version 3 of the License,
# or (at your option) any later version.
# Remotior Sensus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty 
# of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with Remotior Sensus. If not, see <https://www.gnu.org/licenses/>.


import os
import sys
import datetime
import ssl
import smtplib

# sound for Windows OS
try:
    import winsound
except Exception as error:
    str(error)

from remotior_sensus.core import configurations as cfg


class Progress(object):
    process = cfg.process
    step = 0
    message = cfg.message
    percentage = False
    callback = None
    start_time = None
    elapsed_time = None
    previous_step_time = None
    previous_step = 0
    remaining = ''
    ping = 0

    def __init__(
            self, process=None, step=None, message=None, percentage=None,
            callback=None
    ):
        if process is not None:
            self.process = str(process)
        if step is not None:
            self.step = step
        if message is not None:
            self.message = str(message)
        if percentage is not None:
            self.percentage = percentage
        if callback is not None:
            self.callback = callback

    def finish(self):
        """Ends progress and resets."""
        self.remaining = ''
        if self.callback is not None:
            self.callback(
                process=self.process, message='finished', percentage=100,
                elapsed_time=self.elapsed_time, step=100, previous_step=100,
                end=True
            )
        self.process = cfg.process
        self.step = 0
        self.message = cfg.message
        self.percentage = False
        self.start_time = None
        self.elapsed_time = None
        self.previous_step_time = None
        self.previous_step = 0
        finish_sound()
        send_smtp_message(subject=self.process, message='finished')

    # get progress
    def get(self):
        return self.process, self.step, self.message, self.percentage

    # update progress
    def update(
            self, process=None, step=None, message=None, percentage=None,
            start=None, end=None, ping=None, steps=None,
            minimum=None, maximum=None
    ):
        if process is not None:
            self.process = str(process)
        if step is not None:
            # calculate increment
            if minimum is not None and maximum is not None:
                delta = maximum - minimum
                increment = delta / steps
                self.step = int(minimum + step * increment)
            else:
                self.step = step
        if message is not None:
            self.message = str(message)
        if percentage is not None:
            self.percentage = percentage
        if start:
            self.start_time = datetime.datetime.now()
            self.previous_step = 0
            self.step = 0
            self.callback(
                process=self.process, step=self.step, message=self.message,
                percentage=self.percentage, elapsed_time=self.elapsed_time,
                previous_step=self.previous_step, start=start
            )
            return
        if end:
            self.finish()
            return
        step_time = datetime.datetime.now()
        try:
            self.elapsed_time = (step_time - self.start_time).total_seconds()
            progress_time = (
                    step_time - self.previous_step_time).total_seconds()
        except Exception as err:
            str(err)
            self.elapsed_time = None
            progress_time = None
        if ping:
            self.ping = not self.ping
        if self.callback is not None and (
                progress_time is None or progress_time > cfg.refresh_time):
            self.callback(
                process=self.process, step=self.step, message=self.message,
                percentage=self.percentage, elapsed_time=self.elapsed_time,
                previous_step=self.previous_step, start=start, ping=self.ping
            )
        if self.previous_step < self.step:
            self.previous_step = self.step
            if progress_time is None or progress_time > cfg.refresh_time:
                self.previous_step_time = step_time

    # print progress replacing line
    @staticmethod
    def print_progress_replace(
            process=None, step=None, message=None, percentage=None,
            elapsed_time=None, previous_step=None, start=None, end=None,
            ping=0
    ):
        progress_symbols = ['○', '◔', '◑', '◕', '⬤', '⚙']
        colon = [' ', ':']
        if start:
            print('\r', '{} [{}%]{}{}:{} {}'.format(
                    process, str(step).rjust(3, ' '), '', '',
                    message, progress_symbols[-1]
                ), end='\x1b[K'
            )
        elif end:
            if elapsed_time is not None:
                e_time = (' [elapsed {}min{}sec]'.format(
                    int(elapsed_time / 60), str(
                        int(
                            60 * (
                                    (elapsed_time / 60)
                                    - int(elapsed_time / 60))
                        )
                    ).rjust(2, '0')
                ))
            else:
                e_time = ''
            print(
                '\r', '{} [{}%]{}{}:{} {}'.format(
                    process, str(100).rjust(3, ' '), e_time, '',
                    '', progress_symbols[-2]
                ), end='\x1b[K'
            )
        else:
            if not percentage and percentage is not None:
                percentage = -25
            if elapsed_time is not None:
                e_time = (' [elapsed {}min{}sec]'.format(
                    int(elapsed_time / 60), str(
                        int(
                            60 * (
                                    (elapsed_time / 60)
                                    - int(elapsed_time / 60))
                        )
                    ).rjust(2, '0')
                ))
                if previous_step < step:
                    try:
                        remaining_time = (
                                (100 - int(step)) * elapsed_time / int(step))
                        minutes = int(remaining_time / 60)
                        seconds = round(
                            60 * ((remaining_time / 60)
                                  - int(remaining_time / 60))
                        )
                        if seconds == 60:
                            seconds = 0
                            minutes += 1
                        remaining = ' [remaining {}min{}sec]'.format(
                            minutes, str(seconds).rjust(2, '0')
                        )
                        Progress.remaining = remaining
                    except Exception as err:
                        str(err)
                        remaining = ''
                else:
                    remaining = Progress.remaining
            else:
                e_time = ''
                remaining = ''
            try:
                print(
                    '\r', '{} [{}%]{}{}{}{} {}'.format(
                        process, str(step).rjust(3, ' '), e_time, remaining,
                        colon[ping], message,
                        progress_symbols[int(percentage / 25)]
                    ), end='\x1b[K'
                )
            except Exception as err:
                str(err)
                print(str(process))

    # print progress always in a new line
    @staticmethod
    def print_progress(
            process=None, step=None, message=None, percentage=None,
            elapsed_time=None, previous_step=None, start=None, end=None,
            ping=0
    ):
        progress_symbols = ['○', '◔', '◑', '◕', '⬤', '⚙']
        colon = [' ', ':']
        if start:
            print(
                '{} [{}%]{}{}:{} {}'.format(
                    process, str(100).rjust(3, ' '), '', '',
                    message, progress_symbols[-2]
                )
            )
        elif end:
            pass
        else:
            if not percentage and percentage is not None:
                percentage = -25
            if elapsed_time is not None:
                e_time = (' [elapsed {}min{}sec]'.format(
                    int(elapsed_time / 60), str(
                        int(
                            60 * (
                                    (elapsed_time / 60)
                                    - int(elapsed_time / 60))
                        )
                    ).rjust(2, '0')
                ))
                if previous_step < step:
                    try:
                        remaining_time = (
                                (100 - int(step)) * elapsed_time / int(step))
                        minutes = int(remaining_time / 60)
                        seconds = round(
                            60 * ((remaining_time / 60)
                                  - int(remaining_time / 60))
                        )
                        if seconds == 60:
                            seconds = 0
                            minutes += 1
                        remaining = ' [remaining {}min{}sec]'.format(
                            minutes, str(seconds).rjust(2, '0')
                        )
                        Progress.remaining = remaining
                    except Exception as err:
                        str(err)
                        remaining = ''
                else:
                    remaining = Progress.remaining
            else:
                e_time = ''
                remaining = ''
            try:
                print(
                    '{} [{}%]{}{}{}{} {}'.format(
                        process, str(step).rjust(3, ' '), e_time, remaining,
                        colon[ping],
                        message, progress_symbols[int(percentage / 25)]
                    )
                )
            except Exception as err:
                str(err)
                print(str(process))


# beep sound
def beeps(frequency: int, duration: float):
    if sys.platform.startswith('win'):
        winsound.Beep(frequency, int(duration * 1000))
    elif sys.platform.startswith('linux'):
        os.system(
            'play --no-show-progress --null --channels 1 synth %s sine %s'
            % (str(duration), str(frequency))
            )


# finish sound
def finish_sound():
    if cfg.sound_notification is True:
        try:
            beeps(800, 0.2)
            beeps(600, 0.3)
            beeps(700, 0.5)
        except Exception as err:
            str(err)


# send SMTP message
def send_smtp_message(subject: str = None, message: str = None):
    if cfg.smtp_notification is True:
        try:
            if len(cfg.smtp_server) > 0:
                server = smtplib.SMTP(cfg.smtp_server, 587)
                context = ssl.create_default_context()
                server.starttls(context=context)
                server.login(cfg.smtp_user, cfg.smtp_password)
                tolist = cfg.smtp_recipients.split(',')
                if subject is None:
                    subject = 'completed process'
                if message is None:
                    message = 'completed process'
                msg = 'From: %s\nTo: \nSubject: %s\n\n%s' % (
                    cfg.smtp_user, subject, message)
                server.sendmail(cfg.smtp_user, tolist, msg)
                server.quit()
        except Exception as err:
            str(err)



# Remotior Sensus , software to process remote sensing and GIS data.
# Copyright (C) 2022-2023 Luca Congedo.
# Author: Luca Congedo
# Email: ing.congedoluca@gmail.com
#
# This file is part of Remotior Sensus.
# Remotior Sensus is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License,
# or (at your option) any later version.
# Remotior Sensus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty
# of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with Remotior Sensus. If not, see <https://www.gnu.org/licenses/>.
"""
Raster split.

This tool allows for splitting a raster to single bands.

Typical usage example:

    >>> # import Remotior Sensus and start the session
    >>> import remotior_sensus
    >>> rs = remotior_sensus.Session()
    >>> # start the process
    >>> split = rs.raster_split(raster_path='input_path', 
    ... output_path='output_path')
"""  # noqa: E501

from typing import Optional

from remotior_sensus.core import configurations as cfg
from remotior_sensus.core.output_manager import OutputManager
from remotior_sensus.util import (
    files_directories, raster_vector, shared_tools
)


def raster_split(
        raster_path: str, output_path: str = None,
        prefix: Optional[str] = None,
        extent_list: Optional[list] = None,
        n_processes: Optional[int] = None,
        virtual_output: Optional[bool] = None
):
    """Split a multiband raster to single bands.

    This tool allows for splitting a multiband raster to single bands.

    Args:
        raster_path: path of raster used as input.
        output_path: string of output directory path.
        prefix: optional string for output name prefix.
        extent_list: list of boundary coordinates left top right bottom.
        n_processes: number of parallel processes.
        virtual_output: if True (and output_path is directory), save output
            as virtual raster.

    Returns:
        object :func:`~remotior_sensus.core.output_manager.OutputManager` with
            - path = output path

    Examples:
        Perform the split of a raster
            >>> split = raster_split(raster_path='input_path', 
            ... output_path='output_path')
    """  # noqa: E501
    cfg.logger.log.info('start')
    cfg.progress.update(
        process=__name__.split('.')[-1].replace('_', ' '), message='starting',
        start=True
    )
    raster_path = files_directories.input_path(raster_path)
    # prepare process files
    prepared = shared_tools.prepare_process_files(
        input_bands=[raster_path], output_path=output_path,
        n_processes=n_processes, box_coordinate_list=extent_list
    )
    raster_info = prepared['raster_info']
    output_list = []
    bands = raster_info[0][5]
    output_path = output_path.replace('\\', '/').replace('//', '/')
    if output_path.endswith('/'):
        output_path = output_path[:-1]
    if prefix is None:
        prefix = 'band'
    for band in range(bands):
        files_directories.create_parent_directory(output_path)
        out_path = '%s/%s%s' % (output_path, prefix, str(band + 1))
        if virtual_output is True:
            virtual_path = files_directories.output_path(out_path,
                                                         cfg.vrt_suffix)
            output = virtual_path
        else:
            virtual_path = cfg.temp.temporary_file_path(
                name_suffix=cfg.vrt_suffix)
            output = files_directories.output_path(out_path,
                                                   cfg.tif_suffix)
        raster_vector.create_virtual_raster(
            input_raster_list=[raster_path], output=virtual_path,
            band_number_list=[[band + 1]], box_coordinate_list=extent_list,
            relative_to_vrt=False
        )
        if virtual_output is not True:
            raster_vector.gdal_copy_raster(
                input_raster=virtual_path, output=output
                )
        output_list.append(output)
        cfg.progress.update(
            message='splitting', step=band, steps=bands, minimum=1,
            maximum=99, percentage=int(100 * band / bands)
        )
    cfg.progress.update(end=True)
    cfg.logger.log.info('end; raster split: %s' % str(output_list))
    return OutputManager(paths=output_list)



from unittest import TestCase

from remotior_sensus.util import plot_tools


class TestPlotTools(TestCase):

    def test_plot(self):
        pass
        """
        ax = plot_tools.prepare_plot()
        plots, plot_names, x_ticks, y_ticks, v_lines = (
            plot_tools.add_lines_to_plot(
                name_list=['test1', 'test2'],
                wavelength_list=[[0.4, 0.5, 0.6], [0.3, 0.5, 0.7]],
                value_list=[[0.7, 0.4, 0.9], [1.1, 1.3, 1.2]],
                color_list=['red', 'blue'])
        )
        plot_tools.create_plot(
            ax=ax, plots=plots, plot_names=plot_names, x_ticks=x_ticks,
            y_ticks=y_ticks, v_lines=v_lines
        )
        """



# Remotior Sensus , software to process remote sensing and GIS data.
# Copyright (C) 2022-2023 Luca Congedo.
# Author: Luca Congedo
# Email: ing.congedoluca@gmail.com
#
# This file is part of Remotior Sensus.
# Remotior Sensus is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License,
# or (at your option) any later version.
# Remotior Sensus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty
# of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with Remotior Sensus. If not, see <https://www.gnu.org/licenses/>.
"""Cros classification.

This tool performs the cross classification which is similar 
to band combination, but it is executed between two files only.
The reference file can also be of vector type.
A unique value is assigned to each combination of values.
The output is a raster made of unique values corresponding to combinations
of values.
An output text file describes the correspondance between unique values
and combinations, as well as the statistics of each combination,
with the option to calculate an error matrix or linear regression.

Typical usage example:

    >>> # import Remotior Sensus and start the session
    >>> import remotior_sensus
    >>> rs = remotior_sensus.Session()
    >>> # start the process
    >>> cross = rs.cross_classification(classification_path='file1.tif',reference_path='file2.tif',output_path='output.tif')
"""  # noqa: E501

import io
from typing import Optional

import numpy as np

from remotior_sensus.core import configurations as cfg, table_manager as tm
from remotior_sensus.core.output_manager import OutputManager
from remotior_sensus.tools.band_combination import band_combination
from remotior_sensus.util import (
    files_directories, raster_vector, read_write_files, shared_tools
)


def cross_classification(
        classification_path: str, reference_path: str,
        output_path: Optional[str] = None, overwrite: Optional[bool] = False,
        vector_field: Optional[str] = None, nodata_value: Optional[int] = None,
        cross_matrix: Optional[bool] = False,
        regression_raster: Optional[bool] = False,
        error_matrix: Optional[bool] = False,
        extent_list: Optional[list] = None,
        n_processes: Optional[int] = None, available_ram: Optional[int] = None
) -> OutputManager:
    """Calculation of cross classification.

    This tool allows for the cross classification of two files (a
    classification raster and a reference vector or raster) in order to get
    a raster where each value corresponds to a combination of class values.
    Input raster values must be integer type.
    The output is a cross raster and, depending on the parameters, a text
    file reporting the statistics of each combination, error matrix,
    or linear regression statistics.

    Args:
        classification_path: path of raster used as classification input.
        reference_path: path of the vector or raster file used as reference input.
        vector_field: in case of vector reference, the name of the field used as reference value.
        output_path: path of the output raster.
        overwrite: if True, output overwrites existing files.
        nodata_value: value to be considered as nodata.
        cross_matrix: if True then calculate the cross matrix.
        regression_raster: if True then calculate linear regression statistics.
        error_matrix: if True then calculate error matrix.
        extent_list: list of boundary coordinates left top right bottom.
        n_processes: number of parallel processes.
        available_ram: number of megabytes of RAM available to processes.

    Returns:
        Object :func:`~remotior_sensus.core.output_manager.OutputManager` with
            - paths = [output raster path, output table path]

    Examples:
        Perform the cross classification between two files
            >>> cross = cross_classification(classification_path='file1.tif',reference_path='file2.tif',output_path='output.tif')

        Perform the cross classification between two files and calculate the error matrix
            >>> cross = cross_classification(classification_path='file1.tif',reference_path='file2.tif',output_path='output.tif',error_matrix=True)
    """  # noqa: E501
    cfg.logger.log.info('start')
    cfg.progress.update(
        process=__name__.split('.')[-1].replace('_', ' '), message='starting',
        start=True
    )
    # check output path
    out_path, vrt_r = files_directories.raster_output_path(
        output_path,
        overwrite=overwrite
        )
    vector, raster, reference_crs = raster_vector.raster_or_vector_input(
        reference_path
    )
    if extent_list is not None:
        if raster:
            # prepare process files
            prepared = shared_tools.prepare_process_files(
                input_bands=[classification_path, reference_path],
                output_path=output_path,
                overwrite=overwrite, n_processes=n_processes,
                box_coordinate_list=extent_list,
                multiple_output=True, multiple_input=True
            )
            input_raster_list = prepared['input_raster_list']
            n_processes = prepared['n_processes']
            classification_path, reference_path = input_raster_list
        else:
            # prepare process files
            prepared = shared_tools.prepare_process_files(
                input_bands=[classification_path], output_path=output_path,
                overwrite=overwrite, n_processes=n_processes,
                box_coordinate_list=extent_list,
                multiple_output=True, multiple_input=True
            )
            input_raster_list = prepared['input_raster_list']
            n_processes = prepared['n_processes']
            classification_path = input_raster_list[0]
    classification_crs = raster_vector.get_crs(classification_path)
    # check crs
    same_crs = raster_vector.compare_crs(reference_crs, classification_crs)
    # if reference is raster
    if raster:
        if not same_crs:
            t_pmd = cfg.temp.temporary_raster_path(extension=cfg.vrt_suffix)
            reference_raster = cfg.multiprocess.create_warped_vrt(
                raster_path=reference_path, output_path=t_pmd,
                output_wkt=str(classification_crs)
            )
        else:
            reference_raster = reference_path
    # if reference is vector
    else:
        if vector_field is None:
            cfg.logger.log.error('vector field missing')
            cfg.messages.error('vector field missing')
            return OutputManager(check=False)
        if not same_crs:
            # project vector to raster crs
            t_vector = cfg.temp.temporary_file_path(
                name_suffix=cfg.gpkg_suffix
            )
            try:
                raster_vector.reproject_vector(
                    reference_path, t_vector, reference_crs, classification_crs
                )
            except Exception as err:
                cfg.logger.log.error(str(err))
                cfg.messages.error(str(err))
                return OutputManager(check=False)
            reference_path = t_vector
        # convert vector to raster
        reference_raster = cfg.temp.temporary_raster_path(
            extension=cfg.tif_suffix
        )
        # perform conversion
        cfg.multiprocess.multiprocess_vector_to_raster(
            vector_path=reference_path, field_name=vector_field,
            output_path=reference_raster,
            reference_raster_path=classification_path, nodata_value=0,
            background_value=0, available_ram=available_ram,
            minimum_extent=False
        )
    combination = band_combination(
        input_bands=[reference_raster, classification_path],
        output_path=out_path, nodata_value=nodata_value,
        n_processes=n_processes, available_ram=available_ram,
        output_table=False, progress_message=False
    )
    vrt_check = combination.paths[0]
    rec_combinations_array = combination.extra['combinations']
    sum_val = combination.extra['sums']
    cfg.progress.update(message='output table', step=90)
    # get pixel unit
    (gt, crs, un, xy_count, nd, number_of_bands, block_size, scale_offset,
     data_type) = raster_vector.raster_info(out_path)
    p_x = gt[1]
    p_y = abs(gt[5])
    joined_table = tm.join_tables(
        table1=rec_combinations_array, table2=sum_val, field1_name='new_val',
        field2_name='new_val', nodata_value=cfg.nodata_val_Int64,
        join_type='left', progress_message=False
    )
    # create table
    table, slope, intercept, unique_values = _cross_table(
        table=joined_table[joined_table['sum'] != cfg.nodata_val_Int64],
        crs_unit=un, pixel_size_x=p_x, pixel_size_y=p_y,
        regression_raster=regression_raster, cross_matrix=cross_matrix,
        error_matrix=error_matrix
    )
    # save combination to table
    tbl_out = shared_tools.join_path(
        files_directories.parent_directory(out_path), '{}{}'.format(
            files_directories.file_name(out_path, suffix=False), cfg.csv_suffix
        )
    ).replace('\\', '/')
    read_write_files.write_file(table, tbl_out)
    out_raster_b0 = None
    out_raster_b1 = None
    # regression raster
    if regression_raster:
        cfg.progress.update(message='regression raster', step=98)
        # output rasters
        out_raster_b0 = shared_tools.join_path(
            files_directories.parent_directory(out_path), '{}_b0{}'.format(
                files_directories.file_name(out_path, suffix=False),
                cfg.tif_suffix
            )
        ).replace('\\', '/')
        out_raster_b1 = shared_tools.join_path(
            files_directories.parent_directory(out_path), '{}_b1{}'.format(
                files_directories.file_name(out_path, suffix=False),
                cfg.tif_suffix
            )
        ).replace('\\', '/')
        raster_vector.create_raster_from_reference(
            vrt_check, 1, [out_raster_b0], nodata_value=cfg.nodata_val_Float32,
            driver='GTiff', gdal_format='Float32',
            compress=cfg.raster_compression, compress_format='LZW',
            constant_value=intercept
        )
        cfg.progress.update(message='regression raster', step=99)
        raster_vector.create_raster_from_reference(
            vrt_check, 1, [out_raster_b1], nodata_value=cfg.nodata_val_Float32,
            driver='GTiff', gdal_format='Float32',
            compress=cfg.raster_compression, compress_format='LZW',
            constant_value=slope
        )
    cfg.progress.update(end=True)
    cfg.logger.log.info(
        'end; cross classification: %s; table: %s'
        % (str(out_path), str(tbl_out))
    )
    return OutputManager(
        paths=[out_path, tbl_out],
        extra={
            'unique_values': unique_values,
            'regression_raster_b0': out_raster_b0,
            'regression_raster_b1': out_raster_b1
        }
    )


# create text for table
# noinspection PyTypeChecker
def _cross_table(
        table, crs_unit, pixel_size_x, pixel_size_y, regression_raster=False,
        cross_matrix=False, error_matrix=False
):
    cfg.logger.log.debug(
        '_cross_table regression_raster=%s, cross_matrix=%s, error_matrix:%s'
        % (regression_raster, cross_matrix, error_matrix)
    )
    slope = ''
    intercept = ''
    text = []
    cv = cfg.comma_delimiter
    nl = cfg.new_line
    # table
    if 'degree' not in crs_unit:
        output_field_names = ['RasterValue', 'Reference', 'Classification',
                              'PixelSum', 'Area [%s^2]' % crs_unit]
        input_field_names = ['new_val', 'f0', 'f1', 'sum', 'area']
        cross_class = tm.calculate(
            matrix=table, expression_string='"sum" * %s * %s' % (
                str(pixel_size_x), str(pixel_size_y)),
            output_field_name='area', progress_message=False
        )
    else:
        output_field_names = ['RasterValue', 'Reference', 'Classification',
                              'PixelSum']
        input_field_names = ['new_val', 'f0', 'f1', 'sum']
        cross_class = table
    redefined = tm.redefine_matrix_columns(
        matrix=cross_class, input_field_names=input_field_names,
        output_field_names=output_field_names, progress_message=False
    )
    # export matrix creating stream handler
    stream1 = io.StringIO()
    np.savetxt(stream1, redefined, delimiter=cv, fmt='%1.2f')
    matrix_value = stream1.getvalue()
    for c in output_field_names:
        text.append(c)
        text.append(cv)
    text.pop(-1)
    text.append(nl)
    text.append(matrix_value.replace('.00', ''))
    text.append(nl)
    columns = np.unique(
        np.append(cross_class['f0'], cross_class['f1'])
    ).tolist()
    # cross matrix
    if cross_matrix:
        if 'degree' not in crs_unit:
            text.append('CROSS MATRIX [%s^2]' % (str(crs_unit)))
            text.append(nl)
            cross_matrix = tm.pivot_matrix(
                cross_class, row_field='f0', secondary_row_field_list=['f1'],
                column_function_list=[['area', 'sum']], cross_matrix=True,
                progress_message=False
            )
        else:
            text.append('CROSS MATRIX [PixelSum]')
            text.append(nl)
            cross_matrix = tm.pivot_matrix(
                cross_class, row_field='f0', secondary_row_field_list=['f1'],
                column_function_list=[['sum', 'sum']], cross_matrix=True,
                progress_message=False
            )
        text.append(cv)
        text.append('Classification >')
        text.append(nl)
        text.append('Reference V')
        text.append(cv)
        totals = []
        for c in tm.columns(cross_matrix):
            if c != 'f0':
                totals.append(str(int(cross_matrix[c].sum())))
                totals.append(cv)
                text.append(c)
                text.append(cv)
        totals.pop(-1)
        joined_totals = ''.join(totals)
        text.pop(-1)
        text.append(nl)
        # export matrix creating stream handler
        stream2 = io.StringIO()
        np.savetxt(stream2, cross_matrix, delimiter=cv, fmt='%1.2f')
        matrix2_value = stream2.getvalue()
        text.append(matrix2_value.replace('.00', ''))
        text.append('Total')
        text.append(cv)
        text.append(joined_totals)
        text.append(nl)
    elif error_matrix:
        # array matrix
        text.append('ERROR MATRIX [pixel count]')
        text.append(nl)
        text.append(cv)
        text.append('> Reference')
        text.append(nl)
        cross_matrix = tm.pivot_matrix(
            cross_class, row_field='f0', secondary_row_field_list=['f1'],
            column_function_list=[['sum', 'sum']], cross_matrix=True,
            progress_message=False
        )
        # regular error matrix
        err_matrix = np.zeros((len(columns) + 1, len(columns) + 2))
        # area based error matrix
        err_matrix_unbiased = np.zeros((len(columns) + 1, len(columns) + 3))
        text.append('V_Classified')
        text.append(cv)
        text_matrix_unbiased = ['V_Classified', cv]
        # copy values from cross matrix to error matrix
        c = 1
        for column in columns:
            text.append(str(column))
            text.append(cv)
            text_matrix_unbiased.append(str(column))
            text_matrix_unbiased.append(cv)
            if str(column) in tm.columns(cross_matrix):
                err_matrix[:len(columns), c] = cross_matrix[str(column)]
                # copy for accuracy metrics in case area based values are
                # not calculated
                err_matrix_unbiased[:len(columns), c] = cross_matrix[
                    str(column)]
            c += 1
        err_matrix[:len(columns), 0] = np.array(columns)
        text.append('Total')
        text.append(nl)
        total_pixels = err_matrix.sum() - err_matrix[::, 0].sum()
        # add column as total rows N_i
        err_matrix[::, -1] = err_matrix.sum(axis=1) - err_matrix[::, 0]
        # add row of total columns
        err_matrix[-1, ::] = err_matrix.sum(axis=0)
        # nodata value to be masked
        nd = -99999
        err_matrix[-1, 0] = nd
        # area based matrix as proportion P_ij of class over total pixels
        err_matrix_unbiased[::, 0:err_matrix_unbiased.shape[1] - 2] = \
            err_matrix[::, 0:err_matrix.shape[1] - 1] / total_pixels
        err_matrix_unbiased[::, err_matrix_unbiased.shape[1] - 2] = \
            err_matrix[::, err_matrix.shape[1] - 1] * (
                    pixel_size_x * pixel_size_y)
        err_matrix_unbiased[::, err_matrix_unbiased.shape[1] - 1] = \
            err_matrix[::, err_matrix.shape[1] - 1] / total_pixels
        # replace first columns
        err_matrix_unbiased[:-1, 0] = err_matrix[:-1, 0]
        err_matrix_unbiased[-1, 0] = nd
        # export matrix creating stream handler
        stream = io.StringIO()
        # error matrix
        np.savetxt(stream, err_matrix, delimiter=cv, fmt='%i')
        err_matrix_value = stream.getvalue()
        text.append(
            err_matrix_value.replace('.00', '').replace(str(nd), 'Total')
        )
        if 'degree' not in crs_unit:
            text.append(nl)
            # area based error matrix (see Olofsson, et al., 2014,
            # Good practices for estimating area and assessing
            # accuracy of land change, Remote Sensing of Environment, 148,
            # 42-57)
            text.append(
                'AREA BASED ERROR MATRIX%s%s> Reference%s' % (nl, cv, nl)
            )
            # added column Wi is area proportion for class i
            joined_text_matrix_unbiased = ''.join(text_matrix_unbiased)
            text.append(joined_text_matrix_unbiased)
            text.append('Area%sWi%s' % (cv, nl))
            # export matrix creating stream handler
            stream_i = io.StringIO()
            # area based matrix
            np.savetxt(stream_i, err_matrix_unbiased, delimiter=cv, fmt='%.4f')
            err_matrix_unbiased_value = stream_i.getvalue()
            text.append(
                err_matrix_unbiased_value.replace('.0000', '').replace(
                    str(nd), 'Total'
                )
            )
            # estimated area as population proportion P_ij * total_area
            total_area = err_matrix_unbiased[
                         :-1, err_matrix_unbiased.shape[1] - 2].sum()
            estimated_area_matrix = err_matrix_unbiased[-1, 1:-2] * total_area
            estimated_area_matrix = estimated_area_matrix.reshape(
                1, estimated_area_matrix.shape[0]
            )
            # export matrix creating stream handler
            stream_estimated_area = io.StringIO()
            np.savetxt(
                stream_estimated_area, estimated_area_matrix, delimiter=cv,
                fmt='%.2f'
            )
            estimated_area_text = stream_estimated_area.getvalue()
            text.append(
                'Estimated area%s%s%s%s%s' % (
                    cv, estimated_area_text.replace('.00', '').rstrip(nl), cv,
                    str('%1.2f' % total_area).replace('.00', ''), nl)
            )
            # standard error SE_j as summation for each class i of (Wi *
            # P_ij - P_ij^2) / (N_i - 1)
            # create W_i and N_i -1 matrices for calculation
            w_i = np.zeros((len(columns), len(columns)))
            n_i = np.zeros((len(columns), len(columns)))
            for c in range(len(columns)):
                w_i[::, c] = err_matrix_unbiased[:-1, -1]
                n_i[::, c] = err_matrix[:-1, -1] - 1
            std_err_matrix = (err_matrix_unbiased[
                              :-1, 1:-2] * w_i - err_matrix_unbiased[
                                                 :-1, 1:-2] ** 2) / n_i
            std_err = np.sqrt(np.nansum(std_err_matrix, axis=0))
            std_err = std_err.reshape(1, std_err.shape[0])
            # export matrix creating stream handler
            stream_std_err = io.StringIO()
            np.savetxt(stream_std_err, std_err, delimiter=cv, fmt='%.4f')
            std_err_text = stream_std_err.getvalue()
            text.append('SE%s%s' % (cv, std_err_text.replace('.0000', '')))
            # standard error area
            std_err_area = std_err * total_pixels * pixel_size_x * pixel_size_y
            stream_std_err_area = io.StringIO()
            np.savetxt(
                stream_std_err_area, std_err_area, delimiter=cv, fmt='%.2f'
            )
            std_err_area_text = stream_std_err_area.getvalue()
            text.append(
                'SE area%s%s' % (cv, std_err_area_text.replace('.00', ''))
            )
            # confidence interval
            confidence_interval = std_err * total_pixels * (
                    pixel_size_x * pixel_size_y) * 1.96
            stream_confidence_interval = io.StringIO()
            np.savetxt(
                stream_confidence_interval, confidence_interval, delimiter=cv,
                fmt='%.2f'
            )
            confidence_interval_text = stream_confidence_interval.getvalue()
            text.append(
                '95% CI area{}{}'.format(
                    cv, confidence_interval_text.replace('.00', '')
                )
            )
        # accuracy metrics
        nii_tot = 0
        text_p = ['PA [%]']
        text_u = ['UA [%]']
        try:
            for g in range(len(columns)):
                nii = err_matrix_unbiased[g, g + 1]
                nii_tot = nii_tot + nii
                nip = err_matrix_unbiased[g, 1:(len(columns) + 1)].sum()
                npi = err_matrix_unbiased[0:len(columns), g + 1].sum()
                if npi != 0:
                    p = 100 * nii / npi
                else:
                    p = np.nan
                if nip != 0:
                    u = 100 * nii / nip
                else:
                    u = np.nan
                text_p.append(cv)
                text_p.append(str('%1.4f' % p))
                text_u.append(cv)
                text_u.append(str('%1.4f' % u))
        except Exception as err:
            cfg.logger.log.error(str(err))
        joined_text_p = ''.join(text_p)
        joined_text_u = ''.join(text_u)
        text.append(joined_text_p)
        text.append(nl)
        text.append(joined_text_u)
        text.append(nl)
        text.append(nl)
        text.append(
            'Overall accuracy [%] = {}'.format('%1.4f' % (nii_tot * 100))
        )
        text.append(nl)
        text.append(nl)
        text.append('Area unit = %s^2' % crs_unit)
        text.append(nl)
        text.append('SE = standard error')
        text.append(nl)
        text.append('CI = confidence interval')
        text.append(nl)
        text.append("PA = producer's accuracy")
        text.append(nl)
        text.append("UA = user's accuracy")
        text.append(nl)
    # calculate regression
    if regression_raster:
        # linear regression y = b0 + b1 x + E
        r_coefficient = ''
        r_coefficient2 = ''
        var_y = ''
        var_slope = ''
        conf_slope = ''
        var_intercept = ''
        conf_intercept = ''
        # sum_x and sum_y, total, x_mean and y_mean for calculation of sums
        # S_xx, S_yy and S_xy
        r_sum_x = np.nansum(cross_class['f0'] * cross_class['sum'])
        r_sum_y = np.nansum(cross_class['f1'] * cross_class['sum'])
        r_sum_tot = cross_class['sum'].sum()
        r_x_mean = r_sum_x / r_sum_tot
        r_y_mean = r_sum_y / r_sum_tot
        # S_xy = summation from 1 to n of (x - x_mean) * (y - y_mean)
        s_xy = np.nansum(
            (cross_class['f0'] - r_x_mean) * (cross_class['f1'] - r_y_mean) *
            cross_class['sum']
        )
        # S_xx = summation from 1 to n of (x - x_mean)^2
        s_xx = np.nansum(
            ((cross_class['f0'] - r_x_mean) ** 2) * cross_class['sum']
        )
        # S_yy = summation from 1 to n of (y - y_mean)^2
        s_yy = np.nansum(
            ((cross_class['f1'] - r_y_mean) ** 2) * cross_class['sum']
        )
        try:
            r_coefficient = s_xy / (s_xx * s_yy) ** 0.5
            r_coefficient2 = r_coefficient ** 2
            slope = s_xy / s_xx
            intercept = r_y_mean - slope * r_x_mean
            var_y = (s_yy - slope * s_xy) / (r_sum_tot - 2)
            var_slope = var_y / s_xx
            var_intercept = var_y * (1 / r_sum_tot + r_x_mean ** 2 / s_xx)
            conf_slope = 2 * var_slope ** 0.5
            conf_intercept = 2 * var_intercept ** 0.5
        except Exception as err:
            cfg.logger.log.error(str(err))
        text.append(nl)
        text.append('Linear regression Y = B0 + B1*X')
        text.append(nl)
        text.append(
            'Coeff. det. R^2%s%s' % (cfg.tab_delimiter, str(r_coefficient2))
        )
        text.append(nl)
        text.append(
            'Coeff. correlation r%s%s' % (
                cfg.tab_delimiter, str(r_coefficient))
        )
        text.append(nl)
        text.append(
            'B0%s%s ± %s' % (
                cfg.tab_delimiter, str(intercept), str(conf_intercept))
        )
        text.append(nl)
        text.append(
            'B1%s%s ± %s' % (cfg.tab_delimiter, str(slope), str(conf_slope))
        )
        text.append(nl)
        text.append('Variance Y%s%s' % (cfg.tab_delimiter, str(var_y)))
        text.append(nl)
        text.append(
            'Variance B0%s%s' % (cfg.tab_delimiter, str(var_intercept))
        )
        text.append(nl)
        text.append('Variance B1%s%s' % (cfg.tab_delimiter, str(var_slope)))
        text.append(nl)
    joined_text = ''.join(text)
    return joined_text, slope, intercept, columns



from unittest import TestCase

import remotior_sensus
from remotior_sensus.util import files_directories


class TestBandClip(TestCase):

    def test_band_clip(self):
        rs = remotior_sensus.Session(
            n_processes=2, available_ram=1000, log_level=10
            )
        cfg = rs.configurations
        cfg.logger.log.debug('test')
        catalog = rs.bandset_catalog()
        file_list = ['S2_2020-01-01/S2_B02.tif', 'S2_2020-01-01/S2_B03.tif',
                     'S2_2020-01-01/S2_B04.tif']
        root_directory = './data'
        catalog.create_bandset(
            file_list, wavelengths=['Sentinel-2'], bandset_number=1,
            root_directory=root_directory
            )
        cfg.logger.log.debug('>>> test band clip input BandSet')
        # box coordinate list
        extent_list = [230250, 4674510, 230320, 4674440]
        output = rs.band_clip(input_bands=catalog.get_bandset(1),
                              output_path=cfg.temp.dir, prefix='clip_',
                              extent_list=extent_list)
        self.assertTrue(output.check)
        # box coordinate list
        extent_list = [230250, 4674510, 230320, 4674440]
        output = rs.band_clip(input_bands=catalog.get_bandset(1),
                              output_path=cfg.temp.dir, prefix='clip2_',
                              extent_list=extent_list, virtual_output=True)
        self.assertTrue(output.check)
        print('output.paths', output.paths)
        self.assertTrue(files_directories.is_file(output.paths[0]))

        v = './data/files/roi.gpkg'
        output = rs.band_clip(input_bands=catalog.get_bandset(1),
                              output_path=cfg.temp.dir, prefix='clip3_',
                              vector_path=v)
        self.assertTrue(output.check)
        self.assertTrue(files_directories.is_file(output.paths[0]))
        v = './data/files/roi.gpkg'
        output = rs.band_clip(input_bands=catalog.get_bandset(1),
                              output_path=cfg.temp.dir, prefix='clip4_',
                              vector_path=v, vector_field='class')
        self.assertTrue(output.check)
        self.assertTrue(files_directories.is_file(output.paths[0]))

        # clear temporary directory
        #rs.close()



from unittest import TestCase

import remotior_sensus
from remotior_sensus.util import files_directories


class TestBandPCA(TestCase):

    def test_band_pca(self):
        rs = remotior_sensus.Session(
            n_processes=2, available_ram=1000, log_level=10
            )
        cfg = rs.configurations
        cfg.logger.log.debug('test')
        catalog = rs.bandset_catalog()
        file_list = ['S2_2020-01-01/S2_B02.tif', 'S2_2020-01-01/S2_B03.tif',
                     'S2_2020-01-01/S2_B04.tif']
        root_directory = './data'
        catalog.create_bandset(
            file_list, wavelengths=['Sentinel-2'], bandset_number=1,
            root_directory=root_directory
            )
        cfg.logger.log.debug('>>> test band PCA input BandSet')
        temp = cfg.temp.temporary_file_path(name_suffix=cfg.tif_suffix)
        output = rs.band_pca(input_bands=catalog.get_bandset(1),
                             output_path=temp)
        self.assertTrue(output.check)
        self.assertTrue(files_directories.is_file(output.paths[0]))
        cfg.logger.log.debug('>>> test band PCA input BandSet with coordinate')
        coordinate_list = [230250, 4674550, 230320, 4674440]
        temp = cfg.temp.temporary_file_path(name_suffix=cfg.tif_suffix)
        output = rs.band_pca(input_bands=catalog.get_bandset(1),
                             output_path=temp,
                             extent_list=coordinate_list)
        self.assertTrue(output.check)
        self.assertTrue(files_directories.is_file(output.paths[0]))
        cfg.logger.log.debug('>>> test band PCA input BandSet with coordinate')
        bs = catalog.get_bandset(1)
        bs.box_coordinate_list = [230250, 4674550, 230320, 4674440]
        output = rs.band_pca(input_bands=catalog.get_bandset(1),
                             extent_list=coordinate_list)
        self.assertTrue(output.check)
        self.assertTrue(files_directories.is_file(output.paths[0]))

        # clear temporary directory
        rs.close()



# Remotior Sensus , software to process remote sensing and GIS data.
# Copyright (C) 2022-2023 Luca Congedo.
# Author: Luca Congedo
# Email: ing.congedoluca@gmail.com
#
# This file is part of Remotior Sensus.
# Remotior Sensus is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License,
# or (at your option) any later version.
# Remotior Sensus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty
# of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with Remotior Sensus. If not, see <https://www.gnu.org/licenses/>.

"""
Tools to download files
"""

import urllib.request
from datetime import datetime
from http.cookiejar import CookieJar
from os import stat
from time import sleep

from remotior_sensus.core import configurations as cfg
from remotior_sensus.util.files_directories import create_parent_directory
from remotior_sensus.util.read_write_files import write_file


# get proxy handler
def get_proxy_handler(
        proxy_host, proxy_port, proxy_user=None, proxy_password=None
):
    if len(proxy_user) > 0:
        proxy_handler = urllib.request.ProxyHandler(
            {
                'http': 'http://{}:{}@{}:{}'.format(
                    proxy_user, proxy_password, proxy_host, proxy_port
                )
            }
        )
    else:
        proxy_handler = urllib.request.ProxyHandler(
            {'http': 'http://{}:{}'.format(proxy_host, proxy_port)}
        )
    return proxy_handler


# opener
def general_opener(
        proxy_host=None, proxy_port=None, proxy_user=None, proxy_password=None
):
    cookie_jar = CookieJar()
    if proxy_host:
        proxy_handler = get_proxy_handler(
            proxy_host, proxy_port, proxy_user,
            proxy_password
        )
        opener = urllib.request.build_opener(
            proxy_handler, urllib.request.HTTPCookieProcessor(cookie_jar)
        )
    else:
        opener = urllib.request.build_opener(
            urllib.request.HTTPCookieProcessor(cookie_jar)
        )
    return opener, cookie_jar


# open general url
def open_general_url(
        url, proxy_host=None, proxy_port=None, proxy_user=None,
        proxy_password=None
):
    opener, cookie_jar = general_opener(
        proxy_host=proxy_host, proxy_port=proxy_port, proxy_user=proxy_user,
        proxy_password=proxy_password
    )
    request = urllib.request.Request(url)
    try:
        response = opener.open(request)
    except Exception as err:
        response = None
        cfg.logger.log.error('{}; url:{}'.format(err, url))
        cfg.messages.error('{}; url:{}'.format(err, url))
    return response


# download file
def download_file(
        url, output_path, authentication_uri=None, user=None, password=None,
        proxy_host=None, proxy_port=None, proxy_user=None, proxy_password=None,
        progress=True, message=None, min_progress=0, max_progress=100,
        retried=False, timeout=20, callback=None, log=None
):
    cfg.logger.log.debug('url: %s' % url)
    if authentication_uri is None:
        auth_handler = None
    else:
        pass_mgr = urllib.request.HTTPPasswordMgrWithDefaultRealm()
        pass_mgr.add_password(None, authentication_uri, user, password)
        auth_handler = urllib.request.HTTPBasicAuthHandler(pass_mgr)
    cookie_proc = urllib.request.HTTPCookieProcessor()
    # with proxy
    if proxy_host:
        proxy_handler = get_proxy_handler(
            proxy_host, proxy_port, proxy_user, proxy_password
        )
        if auth_handler:
            opener = urllib.request.build_opener(
                proxy_handler, cookie_proc, auth_handler
            )
        else:
            opener = urllib.request.build_opener(proxy_handler, cookie_proc)
    # without proxy
    elif auth_handler:
        opener = urllib.request.build_opener(cookie_proc, auth_handler)
    else:
        opener = urllib.request.build_opener(cookie_proc)
    urllib.request.install_opener(opener)
    try:
        url_request = urllib.request.urlopen(url, timeout=timeout)
        if log is not None:
            log.debug('url_request: %s' % str(url_request))
        try:
            file_size = int(url_request.headers['Content-Length'])
            # size megabyte
            total_size = round(file_size / 1048576, 2)
        except Exception as err:
            str(err)
            file_size = 1
            total_size = 1
        block_size = 1024 * 1024
        # set initial speed for adaptive block size
        speed = 0
        adaptive_block_size = True
        if log is not None:
            log.debug(
                'block_size: %s; file_size: %s'
                % (str(block_size), str(file_size))
                )
        # small files
        if block_size >= file_size:
            response = url_request.read()
            write_file(response, output_path, mode='wb')
        else:
            create_parent_directory(output_path)
            with open(output_path, 'wb') as file:
                while True:
                    if cfg.action is True:
                        if adaptive_block_size:
                            start_time = datetime.now()
                        block_read = url_request.read(block_size)
                        if not block_read:
                            break
                        # adapt block size
                        if adaptive_block_size:
                            try:
                                end_time = datetime.now()
                                time_delta = end_time - start_time
                                new_speed = (
                                        block_size / time_delta.microseconds
                                )
                                if new_speed >= speed:
                                    block_size = block_size + 1024 * 1024
                                    start_time = end_time
                                    speed = new_speed
                            except Exception as err:
                                str(err)
                        if progress is True:
                            try:
                                downloaded_part_size = round(
                                    int(stat(output_path).st_size) / 1048576, 2
                                )
                                if message is None:
                                    message = '({}/{} MB) {}'.format(
                                        downloaded_part_size, total_size, url
                                    )
                                step = int(
                                    (max_progress - min_progress)
                                    * downloaded_part_size / total_size
                                    + min_progress
                                )
                                percentage = int(
                                    100 * downloaded_part_size / total_size
                                )
                                if callback is None:
                                    cfg.progress.update(
                                        message=message, step=step,
                                        percentage=percentage, ping=True
                                    )
                                else:
                                    callback(percentage, False)
                            except Exception as err:
                                str(err)
                        # write file
                        file.write(block_read)
                    else:
                        return False, 'cancel'
        return True, output_path
    except Exception as err:
        if retried is False and '403' not in str(err):
            sleep(2)
            download_file(
                url=url, output_path=output_path,
                authentication_uri=authentication_uri, user=user,
                password=password,
                proxy_host=proxy_host, proxy_port=proxy_port,
                proxy_user=proxy_user,
                proxy_password=proxy_password,
                progress=progress, message=message, min_progress=min_progress,
                max_progress=max_progress, retried=True, timeout=timeout * 2
            )
        else:
            return False, str(err)
